import datetime
import json
import os
import sys
import time
from dataclasses import dataclass
from random import randint
from threading import Lock, Thread
import traceback
from typing import Any, List

import numpy as np
import pandas as pd
import torch
import triton_python_backend_utils as pb_utils
from torch import from_numpy
from torch.utils.dlpack import from_dlpack

import tensorrt_llm.bindings.executor as trtllm
from tensorrt_llm.llmapi.tokenizer import _xgrammar_tokenizer_info

METRIC_TOTAL_OUTPUT_TOKENS = "total_output_tokens"
METRIC_TOTAL_INPUT_TOKENS = "total_input_tokens"
import tensorrt_llm.logger as logger

# From https://github.com/pytorch/pytorch/blob/39425feac799905402abe4d15667fa47c344f2d7/torch/testing/_internal/common_utils.py#L1761
# Dict of NumPy dtype -> torch dtype (when the correspondence exists)
numpy_to_torch_dtype_dict = {
    np.bool_: torch.bool,
    np.uint8: torch.uint8,
    np.uint16: torch.uint16,
    np.uint32: torch.uint32,
    np.uint64: torch.uint64,
    np.int8: torch.int8,
    np.int16: torch.int16,
    np.int32: torch.int32,
    np.int64: torch.int64,
    np.float16: torch.float16,
    np.float32: torch.float32,
    np.float64: torch.float64,
    np.complex64: torch.complex64,
    np.complex128: torch.complex128,
}

# Dict of torch dtype -> NumPy dtype
torch_to_numpy_dtype_dict = {value: key for (key, value) in numpy_to_torch_dtype_dict.items()}
torch_to_numpy_dtype_dict.update({torch.bfloat16: np.float32, torch.complex32: np.complex64})


@dataclass
class RequestData:
    triton_req_id: int
    triton_user_id: str
    batch_index: int
    batch_size: int
    num_return_sequences: int
    num_input_tokens: int
    num_output_tokens: int
    response_sender: Any


def mpi_comm():
    from mpi4py import MPI

    return MPI.COMM_WORLD


def mpi_rank():
    return mpi_comm().Get_rank()


def get_input_tensor_by_name(
    request, name, expected_batch_size=None, batch_index=None, force_on_torch=False
):
    tensor = pb_utils.get_input_tensor_by_name(request, name)
    if tensor is None:
        return None

    if tensor.is_cpu() and not force_on_torch:
        tensor = tensor.as_numpy()
    else:
        tensor = from_dlpack(tensor.to_dlpack())

    if expected_batch_size is not None and tensor.shape[0] != expected_batch_size:
        raise pb_utils.TritonModelException(
            f"Expected batch size doesn't match batch size for tensor {name}. Expected {expected_batch_size} got {tensor.shape[0]}"
        )

    if (
        batch_index is not None
        and expected_batch_size is not None
        and batch_index >= expected_batch_size
    ):
        raise pb_utils.TritonModelException(
            f"Invalid batch index in get_input_tensor_by_name for {name}"
        )

    if batch_index is not None:
        # Add leading 1 batch dimension
        if isinstance(tensor, np.ndarray):
            return np.expand_dims(tensor[batch_index], axis=0)
        elif isinstance(tensor, torch.Tensor):
            return torch.unsqueeze(tensor[batch_index], dim=0)
    else:
        return tensor


def get_input_scalar_by_name(request, name, expected_batch_size=1, batch_index=0):
    tensor = pb_utils.get_input_tensor_by_name(request, name)
    if tensor is None:
        return None
    tensor = tensor.as_numpy()

    if tensor.size != expected_batch_size:
        raise pb_utils.TritonModelException(f"Expected a scalar tensor for tensor {name}")

    return tensor.item(batch_index)


def read_parameter_as_type(value, name, pytype=str):
    if value == "":
        return None
    if value.startswith("${") and value.endswith("}"):
        return None
    if pytype is bool:
        return value.lower() in ["1", "true"]
    try:
        result = pytype(value)
        return result
    except:
        pb_utils.Logger.log_warning(
            f"Could not read parameter '{name}' with value '{value}', will use default."
        )
        return None


def get_parameter(model_config, name, pytype=str):
    if name not in model_config["parameters"]:
        return None
    return read_parameter_as_type(model_config["parameters"][name]["string_value"], name, pytype)


def convert_word_list(word_list):
    if word_list is None:
        return None
    word_list = word_list.tolist()
    if len(word_list) == 0 or len(word_list[0]) != 2:
        raise pb_utils.TritonModelException(f"Invalid format for word list.")
    words, indices = word_list[0]
    result = []
    current_index = 0
    for i in indices:
        if i == -1:
            continue
        if i > len(words):
            raise pb_utils.TritonModelException(f"Invalid format for word list.")
        current_word = []
        while current_index < i:
            current_word.append(words[current_index])
            current_index += 1
        result.append(current_word)
    return result


def parse_medusa_choices(medusa_choices):
    if medusa_choices is None:
        return None
    try:
        result = json.loads("[" + medusa_choices.replace("{", "[").replace("}", "]") + "]")
        assert isinstance(result, list) and len(result) > 0
        assert all([isinstance(x, list) for x in result])
        assert all([isinstance(y, int) for x in result for y in x])
    except Exception:
        raise pb_utils.TritonModelException("Invalid format for medusa_choices")
    return result


def parse_eagle_choices(eagle_choices):
    return parse_medusa_choices(eagle_choices)


def get_sampling_config_from_request(request, batch_size=1, batch_index=0):
    kwargs = {}
    kwargs["beam_width"] = (
        get_input_scalar_by_name(request, "beam_width", batch_size, batch_index) or 1
    )
    kwargs["top_k"] = get_input_scalar_by_name(request, "runtime_top_k", batch_size, batch_index)
    kwargs["top_p"] = get_input_scalar_by_name(request, "runtime_top_p", batch_size, batch_index)
    kwargs["top_p"] = None if kwargs["top_p"] is None or kwargs["top_p"] <= 0 else kwargs["top_p"]
    kwargs["random_seed"] = get_input_scalar_by_name(
        request, "random_seed", batch_size, batch_index
    )
    kwargs["temperature"] = get_input_scalar_by_name(
        request, "temperature", batch_size, batch_index
    )
    kwargs["min_length"] = get_input_scalar_by_name(request, "min_length", batch_size, batch_index)
    kwargs["repetition_penalty"] = get_input_scalar_by_name(
        request, "repetition_penalty", batch_size, batch_index
    )
    kwargs["presence_penalty"] = get_input_scalar_by_name(
        request, "presence_penalty", batch_size, batch_index
    )
    kwargs["frequency_penalty"] = get_input_scalar_by_name(
        request, "frequency_penalty", batch_size, batch_index
    )
    kwargs["length_penalty"] = get_input_scalar_by_name(
        request, "len_penalty", batch_size, batch_index
    )
    kwargs["top_p_min"] = get_input_scalar_by_name(
        request, "runtime_top_p_min", batch_size, batch_index
    )
    kwargs["top_p_reset_ids"] = get_input_scalar_by_name(
        request, "runtime_top_p_reset_ids", batch_size, batch_index
    )
    kwargs["top_p_decay"] = get_input_scalar_by_name(
        request, "runtime_top_p_decay", batch_size, batch_index
    )
    kwargs["beam_search_diversity_rate"] = get_input_scalar_by_name(
        request, "beam_search_diversity_rate", batch_size, batch_index
    )
    kwargs["early_stopping"] = get_input_scalar_by_name(
        request, "early_stopping", batch_size, batch_index
    )
    kwargs["num_return_sequences"] = (
        get_input_scalar_by_name(request, "num_return_sequences", batch_size, batch_index) or 1
    )
    kwargs = {k: v for k, v in kwargs.items() if v is not None}
    return trtllm.SamplingConfig(**kwargs)


def get_output_config_from_request(request, batch_size=1, batch_index=0):
    kwargs = {}
    kwargs["return_log_probs"] = get_input_scalar_by_name(
        request, "return_log_probs", batch_size, batch_index
    )
    kwargs["return_context_logits"] = get_input_scalar_by_name(
        request, "return_context_logits", batch_size, batch_index
    )
    kwargs["return_generation_logits"] = get_input_scalar_by_name(
        request, "return_generation_logits", batch_size, batch_index
    )
    kwargs["return_perf_metrics"] = get_input_scalar_by_name(
        request, "return_perf_metrics", batch_size, batch_index
    )
    if get_input_scalar_by_name(request, "return_kv_cache_reuse_stats", batch_size, batch_index):
        pb_utils.Logger.log_warn(
            "return_kv_cache_reuse_stats is deprecated, please use return_perf_metrics instead."
        )
    kwargs = {k: v for k, v in kwargs.items() if v is not None}
    return trtllm.OutputConfig(**kwargs)


def get_external_draft_tokens_config_from_request(request, batch_size=1, batch_index=0):
    kwargs = {}
    draft_input_ids = get_input_tensor_by_name(request, "draft_input_ids", batch_size, batch_index)
    if draft_input_ids is not None:
        kwargs["tokens"] = draft_input_ids[0].tolist()
    draft_logits = get_input_tensor_by_name(request, "draft_logits", batch_size, batch_index)
    if draft_logits is not None:
        kwargs["logits"] = from_numpy(draft_logits).squeeze(dim=0)
    kwargs["acceptance_threshold"] = get_input_scalar_by_name(
        request, "draft_acceptance_threshold", batch_size, batch_index
    )
    kwargs = {k: v for k, v in kwargs.items() if v is not None}
    if len(kwargs) > 0:
        return trtllm.ExternalDraftTokensConfig(**kwargs)
    return None


def get_prompt_tuning_config_from_request(request, batch_size=1, batch_index=0, input_length=0):
    # prompt_vocab_size is unused by executor.
    kwargs = {}
    prompt_embedding_table = get_input_tensor_by_name(
        request, "prompt_embedding_table", batch_size, batch_index
    )
    prompt_table_extra_ids = get_input_tensor_by_name(
        request, "prompt_table_extra_ids", batch_size, batch_index
    )
    if prompt_embedding_table is not None:
        if isinstance(prompt_embedding_table, np.ndarray):
            kwargs["embedding_table"] = from_numpy(prompt_embedding_table).squeeze(dim=0)
        elif isinstance(prompt_embedding_table, torch.Tensor):
            kwargs["embedding_table"] = prompt_embedding_table.squeeze(dim=0)

        if prompt_table_extra_ids is not None:
            prompt_table_extra_ids = prompt_table_extra_ids[0].tolist()
            if len(prompt_table_extra_ids) != 0:
                kwargs["input_token_extra_ids"] = prompt_table_extra_ids[0:input_length]
    kwargs = {k: v for k, v in kwargs.items() if v is not None}
    if len(kwargs) > 0:
        return trtllm.PromptTuningConfig(**kwargs)
    return None


def get_lora_config_from_request(request, batch_size=1, batch_index=0):
    kwargs = {}
    kwargs["task_id"] = get_input_scalar_by_name(request, "lora_task_id", batch_size, batch_index)
    lora_weights = get_input_tensor_by_name(request, "lora_weights", batch_size, batch_index)
    if lora_weights is not None:
        kwargs["weights"] = from_numpy(lora_weights).squeeze(dim=0)
    lora_config = get_input_tensor_by_name(request, "lora_config", batch_size, batch_index)
    if lora_config is not None:
        kwargs["config"] = from_numpy(lora_config).squeeze(dim=0)
    kwargs = {k: v for k, v in kwargs.items() if v is not None}
    if len(kwargs) > 0:
        return trtllm.LoraConfig(**kwargs)
    return None


def get_guided_decoding_params_from_request(request, batch_size=1, batch_index=0):
    kwargs = {}
    guided_decoding_guide_type = get_input_tensor_by_name(
        request, "guided_decoding_guide_type", batch_size, batch_index
    )
    if guided_decoding_guide_type is not None:
        guided_decoding_guide_type = guided_decoding_guide_type.squeeze(axis=0)[0].decode()
        guided_decoding_guide_type_mapping = {
            "json": trtllm.GuidedDecodingParams.GuideType.JSON,
            "json_schema": trtllm.GuidedDecodingParams.GuideType.JSON_SCHEMA,
            "regex": trtllm.GuidedDecodingParams.GuideType.REGEX,
            "ebnf_grammar": trtllm.GuidedDecodingParams.GuideType.EBNF_GRAMMAR,
        }
        guided_decoding_guide_type = guided_decoding_guide_type_mapping.get(
            guided_decoding_guide_type
        )
    kwargs["guide_type"] = guided_decoding_guide_type

    guided_decoding_guide = get_input_tensor_by_name(
        request, "guided_decoding_guide", batch_size, batch_index
    )
    if guided_decoding_guide is not None:
        kwargs["guide"] = guided_decoding_guide.squeeze(axis=0)[0].decode()
    kwargs = {k: v for k, v in kwargs.items() if v is not None}
    if len(kwargs) > 0:
        return trtllm.GuidedDecodingParams(**kwargs)
    return None


def get_kv_cache_retention_config_from_request(request, batch_size=1, batch_index=0):
    def get_tensor_and_check_length(name: str, expected_length: int):
        tensor = get_input_tensor_by_name(request, name, batch_size, batch_index)

        if tensor is None:
            raise RuntimeError(f"{name} must be provided.")

        tensor = np.squeeze(tensor, axis=0)

        if len(tensor) != expected_length:
            raise RuntimeError(
                f"Invalid {name} length. Expected length {expected_length}, got length {len(tensor)}"
            )

        return tensor

    token_range_starts = get_input_tensor_by_name(
        request, "retention_token_range_starts", batch_size, batch_index
    )

    if token_range_starts is not None:
        token_range_starts = np.squeeze(token_range_starts, axis=0)

        token_range_ends = get_tensor_and_check_length(
            "retention_token_range_ends", len(token_range_starts)
        )
        token_range_ends = [None if end == -1 else end for end in token_range_ends]

        token_range_priorities = get_tensor_and_check_length(
            "retention_token_range_priorities", len(token_range_starts)
        )

        token_range_durations_ms = get_input_tensor_by_name(
            request, "retention_token_range_durations_ms", batch_size, batch_index
        )

        if token_range_durations_ms is None:
            token_range_durations_ms = [None] * len(token_range_starts)
        else:
            token_range_durations_ms = np.squeeze(token_range_durations_ms, axis=0)
            token_range_durations_ms = [
                None if duration == -1 else duration for duration in token_range_durations_ms
            ]

            if len(token_range_durations_ms) != len(token_range_starts):
                raise RuntimeError(
                    f"Invalid retention_token_range_durations length. Expected length {len(token_range_starts)}, got length {len(token_range_durations_ms)}"
                )

        ranges = []

        for start, end, priority, duration_ms in zip(
            token_range_starts, token_range_ends, token_range_priorities, token_range_durations_ms
        ):
            ranges.append(
                trtllm.KvCacheRetentionConfig.TokenRangeRetentionConfig(
                    token_start=start,
                    token_end=end,
                    priority=priority.item(),
                    duration_ms=None
                    if duration_ms is None
                    else datetime.timedelta(milliseconds=duration_ms.item()),
                )
            )

        decode_args = {}

        decode_priority = get_input_scalar_by_name(
            request, "retention_decode_priority", batch_size, batch_index
        )
        if decode_priority is not None:
            decode_args["decode_retention_priority"] = decode_priority

        decode_duration_ms = get_input_scalar_by_name(
            request, "retention_decode_duration_ms", batch_size, batch_index
        )
        if decode_duration_ms is not None:
            decode_args["decode_duration_ms"] = (
                decode_duration_ms if decode_duration_ms != -1 else None
            )

        return trtllm.KvCacheRetentionConfig(token_range_retention_configs=ranges, **decode_args)

    return None


def get_lookahead_decoding_config_from_request(
    request, executor_lookahead_config, batch_size=1, batch_index=0
):
    lookahead_window_size = get_input_tensor_by_name(
        request, "lookahead_window_size", batch_size, batch_index
    )

    lookahead_ngram_size = get_input_tensor_by_name(
        request, "lookahead_ngram_size", batch_size, batch_index
    )

    lookahead_verification_set_size = get_input_tensor_by_name(
        request, "lookahead_verification_set_size", batch_size, batch_index
    )

    # None lookahead config for requests.
    if all(
        x is None
        for x in [lookahead_window_size, lookahead_ngram_size, lookahead_verification_set_size]
    ):
        return None

    # Have request lookahead config but no executor config.
    if executor_lookahead_config is None:
        raise RuntimeError(
            "The request lookahead decoding input tensors (window_size, ngram_size and verification_set_size) can only be set if the model instance lookahead parameters are also specified"
        )

    return trtllm.LookaheadDecodingConfig(
        lookahead_window_size, lookahead_ngram_size, lookahead_verification_set_size
    )


def get_mrope_config_from_request(request, batch_size=1, batch_index=0):
    mrope_rotary_cos_sin = get_input_tensor_by_name(
        request, "mrope_rotary_cos_sin", batch_size, batch_index
    )
    mrope_position_deltas = get_input_tensor_by_name(
        request, "mrope_position_deltas", batch_size, batch_index, force_on_torch=False
    )
    assert (mrope_rotary_cos_sin is None) == (
        mrope_position_deltas is None
    ), "Both mrope_rotary_cos_sin and mrope_position_detals must be either None or not None."

    if mrope_rotary_cos_sin is not None and mrope_position_deltas is not None:
        mrope_config = trtllm.MropeConfig(
            mrope_rotary_cos_sin=mrope_rotary_cos_sin[0],
            mrope_position_deltas=mrope_position_deltas[0],
        )
        return mrope_config
    return None


def build_1_2_5_buckets(max_value: int) -> List[int]:
    """
    Builds a list of buckets with increasing powers of 10 multiplied by
    mantissa values (1, 5), starting from 10 until the value exceeds
    the specified maximum.

    Example:
    >>> build_1_2_5_buckets(1000)
    [10, 50, 100, 500, 1000]
    """
    mantissa_lst = [1, 5]
    exponent = 1  # Start from exponent 1 instead of 0
    buckets: List[int] = []
    while True:
        for m in mantissa_lst:
            value = m * 10**exponent
            if value <= max_value:
                buckets.append(value)
            else:
                return buckets
        exponent += 1


def convert_request(request, exclude_input_from_output, decoupled, executor_lookahead_config=None):
    inputs = {}
    input_token_ids = get_input_tensor_by_name(request, "decoder_input_ids")
    if input_token_ids is None:
        raise pb_utils.TritonModelException("A value is required for input_ids")
    if len(input_token_ids.shape) != 2:
        raise pb_utils.TritonModelException(f"Invalid format for input_ids")
    batch_size = input_token_ids.shape[0]
    requests = []
    for batch_index in range(0, batch_size):
        input_token_ids = get_input_tensor_by_name(request, "decoder_input_ids", batch_size, batch_index)[0]
        if input_token_ids is None:
            raise pb_utils.TritonModelException("A value is required for input_ids")
        input_token_ids = input_token_ids.tolist()
        if len(input_token_ids) == 0:
            raise pb_utils.TritonModelException(f"Invalid format for input_ids")

        input_length = get_input_scalar_by_name(request, "input_lengths", batch_size, batch_index)
        if input_length is None:
            input_length = len(input_token_ids)
        # Trim input token ids with input_lengths
        inputs["input_token_ids"] = input_token_ids[0:input_length]
        inputs["max_new_tokens"] = get_input_scalar_by_name(
            request, "request_output_len", batch_size, batch_index
        )
        if inputs["max_new_tokens"] is None:
            raise pb_utils.TritonModelException("A value is required for request_output_len")
        inputs["streaming"] = get_input_scalar_by_name(
            request, "streaming", batch_size, batch_index
        )
        if inputs["streaming"] and not decoupled:
            raise pb_utils.TritonModelException("Streaming is only supported in decoupled mode.")

        inputs["end_id"] = get_input_scalar_by_name(request, "end_id", batch_size, batch_index)
        inputs["pad_id"] = get_input_scalar_by_name(request, "pad_id", batch_size, batch_index)
        inputs["stop_words"] = convert_word_list(
            get_input_tensor_by_name(request, "stop_words_list", batch_size, batch_index)
        )
        inputs["bad_words"] = convert_word_list(
            get_input_tensor_by_name(request, "bad_words_list", batch_size, batch_index)
        )
        embedding_bias = get_input_tensor_by_name(
            request, "embedding_bias", batch_size, batch_index
        )
        if embedding_bias is not None and embedding_bias.size != 0:
            inputs["embedding_bias"] = from_numpy(embedding_bias).squeeze(dim=0)

        sampling_config = get_sampling_config_from_request(request, batch_size, batch_index)
        output_config = get_output_config_from_request(request, batch_size, batch_index)
        req_exclude_input_from_output = get_input_scalar_by_name(
            request, "exclude_input_in_output", batch_size, batch_index
        )
        if req_exclude_input_from_output is None:
            # if request doesn't specify exclude_input_from_output, try to use the parameter
            output_config.exclude_input_from_output = (
                exclude_input_from_output if exclude_input_from_output is not None else False
            )
        else:
            output_config.exclude_input_from_output = req_exclude_input_from_output

        external_draft_tokens_config = get_external_draft_tokens_config_from_request(
            request, batch_size, batch_index
        )
        prompt_tuning_config = get_prompt_tuning_config_from_request(
            request, batch_size, batch_index, input_length
        )
        mrope_config = get_mrope_config_from_request(request, batch_size, batch_index)
        lora_config = get_lora_config_from_request(request, batch_size, batch_index)
        kv_cache_retention_config = get_kv_cache_retention_config_from_request(
            request, batch_size, batch_index
        )
        request_lookahead_config = get_lookahead_decoding_config_from_request(
            request, executor_lookahead_config, batch_size, batch_index
        )

        # Inputs for mllama support
        encoder_input_features = get_input_tensor_by_name(
            request, "encoder_input_features", batch_size, batch_index
        )
        if encoder_input_features is not None:
            if isinstance(encoder_input_features, np.ndarray):
                encoder_input_features = from_numpy(encoder_input_features).squeeze(dim=0)
            elif isinstance(encoder_input_features, torch.Tensor):
                encoder_input_features = encoder_input_features.squeeze(dim=0)
            inputs["encoder_input_features"] = encoder_input_features
            logger.debug(f"inputs to llm: encoder_input_features ({encoder_input_features.shape}")

            encoder_output_length = get_input_tensor_by_name(
                request, "encoder_output_lengths", batch_size, batch_index
            )
            if encoder_output_length is not None:
                inputs["encoder_output_length"] = np.squeeze(encoder_output_length, axis=0)

            cross_attention_mask = get_input_tensor_by_name(
                request, "cross_attention_mask", batch_size, batch_index
            )
            if cross_attention_mask is not None:
                inputs["cross_attention_mask"] = cross_attention_mask[0]
                logger.debug(f"inputs to llm: cross_attention_mask ({ cross_attention_mask.shape})")

            skip_cross_attn_blocks = get_input_tensor_by_name(
                request, "skip_cross_attn_blocks", batch_size, batch_index, force_on_torch=True
            )
            if skip_cross_attn_blocks is not None:
                inputs["skip_cross_attn_blocks"] = skip_cross_attn_blocks[0]
                logger.debug(
                    f"inputs to llm: skip_cross_attn_blocks ({ skip_cross_attn_blocks.shape})"
                )

        guided_decoding_params = get_guided_decoding_params_from_request(
            request, batch_size, batch_index
        )

        requests.append(
            trtllm.Request(
                **inputs,
                sampling_config=sampling_config,
                output_config=output_config,
                external_draft_tokens_config=external_draft_tokens_config,
                prompt_tuning_config=prompt_tuning_config,
                mrope_config=mrope_config,
                lora_config=lora_config,
                guided_decoding_params=guided_decoding_params,
                lookahead_config=request_lookahead_config,
                kv_cache_retention_config=kv_cache_retention_config,
            )
        )
    return requests


def convert_response(
    response, batch_index, batch_size, num_return_sequences, expected_logits_dtype=torch.float32
):
    if response.has_error():
        return (
            pb_utils.InferenceResponse(
                output_tensors=[], error=pb_utils.TritonError(response.error_msg)
            ),
            True,
            0,
        )
    result = response.result
    beam_lengths = np.expand_dims(
        np.array([len(beam) for beam in result.output_token_ids], np.int32), 0
    )
    max_beam_length = max([len(beam) for beam in result.output_token_ids])
    output_ids = np.full((1, len(result.output_token_ids), max_beam_length), -1, np.int32)
    for idx, beam in enumerate(result.output_token_ids):
        output_ids[0, idx, : len(beam)] = beam

    output_lengths = output_ids.size
    output_tensors = [
        pb_utils.Tensor("output_ids", output_ids),
        pb_utils.Tensor("sequence_length", beam_lengths),
    ]

    if result.cum_log_probs is not None:
        output_tensors.append(
            pb_utils.Tensor(
                "cum_log_probs", np.expand_dims(np.array(result.cum_log_probs, np.float32), 0)
            )
        )

    if result.log_probs is not None:
        output_tensors.append(
            pb_utils.Tensor(
                "output_log_probs", np.expand_dims(np.array(result.log_probs, np.float32), 0)
            )
        )

    if result.context_logits is not None:
        assert result.context_logits.dtype is expected_logits_dtype
        output_tensors.append(
            pb_utils.Tensor(
                "context_logits",
                np.expand_dims(
                    np.array(
                        result.context_logits,
                        torch_to_numpy_dtype_dict[result.context_logits.dtype],
                    ),
                    0,
                ),
            )
        )

    if result.generation_logits is not None:
        assert result.generation_logits.dtype is expected_logits_dtype
        output_tensors.append(
            pb_utils.Tensor(
                "generation_logits",
                np.expand_dims(
                    np.array(
                        result.generation_logits,
                        torch_to_numpy_dtype_dict[result.generation_logits.dtype],
                    ),
                    0,
                ),
            )
        )

    if batch_size > 1:
        output_tensors.append(
            pb_utils.Tensor("batch_index", np.expand_dims(np.array([batch_index], np.int32), 0))
        )

    if num_return_sequences > 1:
        output_tensors.append(
            pb_utils.Tensor(
                "sequence_index", np.expand_dims(np.array([result.sequence_index], np.int32), 0)
            )
        )

    if result.request_perf_metrics is not None:
        kv_cache_metrics = result.request_perf_metrics.kv_cache_metrics
        output_tensors.append(
            pb_utils.Tensor(
                "kv_cache_alloc_new_blocks",
                np.expand_dims(np.array([kv_cache_metrics.num_new_allocated_blocks], np.int32), 0),
            )
        )
        output_tensors.append(
            pb_utils.Tensor(
                "kv_cache_reused_blocks",
                np.expand_dims(np.array([kv_cache_metrics.num_reused_blocks], np.int32), 0),
            )
        )
        output_tensors.append(
            pb_utils.Tensor(
                "kv_cache_alloc_total_blocks",
                np.expand_dims(
                    np.array([kv_cache_metrics.num_total_allocated_blocks], np.int32), 0
                ),
            )
        )

        timing_metrics = result.request_perf_metrics.timing_metrics
        output_tensors.append(
            pb_utils.Tensor(
                "arrival_time_ns",
                np.expand_dims(
                    np.array([pd.Timedelta(timing_metrics.arrival_time).value], np.int64), 0
                ),
            )
        )
        output_tensors.append(
            pb_utils.Tensor(
                "first_scheduled_time_ns",
                np.expand_dims(
                    np.array([pd.Timedelta(timing_metrics.first_scheduled_time).value], np.int64), 0
                ),
            )
        )
        output_tensors.append(
            pb_utils.Tensor(
                "first_token_time_ns",
                np.expand_dims(
                    np.array([pd.Timedelta(timing_metrics.first_token_time).value], np.int64), 0
                ),
            )
        )
        output_tensors.append(
            pb_utils.Tensor(
                "last_token_time_ns",
                np.expand_dims(
                    np.array([pd.Timedelta(timing_metrics.last_token_time).value], np.int64), 0
                ),
            )
        )

        spec_dec_metrics = result.request_perf_metrics.speculative_decoding
        output_tensors.append(
            pb_utils.Tensor(
                "acceptance_rate",
                np.expand_dims(np.array([spec_dec_metrics.acceptance_rate], np.float32), 0),
            )
        )
        output_tensors.append(
            pb_utils.Tensor(
                "total_accepted_draft_tokens",
                np.expand_dims(
                    np.array([spec_dec_metrics.total_accepted_draft_tokens], np.int32), 0
                ),
            )
        )
        output_tensors.append(
            pb_utils.Tensor(
                "total_draft_tokens",
                np.expand_dims(np.array([spec_dec_metrics.total_draft_tokens], np.int32), 0),
            )
        )

    return pb_utils.InferenceResponse(output_tensors), result.is_final, output_lengths


def convert_scheduler_policy(batch_scheduler_policy: str):
    if batch_scheduler_policy.lower() == "max_utilization":
        return trtllm.CapacitySchedulerPolicy.MAX_UTILIZATION
    elif batch_scheduler_policy.lower() == "guaranteed_no_evict":
        return trtllm.CapacitySchedulerPolicy.GUARANTEED_NO_EVICT
    raise pb_utils.TritonModelException(
        f"batch_scheduler_policy value of '{batch_scheduler_policy}' is not supported."
    )


def convert_batching_type(gpt_model_type: str):
    if gpt_model_type is None:
        return None
    if (
        gpt_model_type.lower() == "inflight_fused_batching"
        or gpt_model_type.lower() == "inflight_batching"
    ):
        return trtllm.BatchingType.INFLIGHT
    elif gpt_model_type.lower() == "v1":
        return trtllm.BatchingType.STATIC
    raise pb_utils.TritonModelException(
        f"gpt_model_type value of '{gpt_model_type}' is not supported."
    )


def convert_decoding_mode(decoding_mode: str):
    if decoding_mode is None:
        return None
    elif decoding_mode == "auto":
        return trtllm.DecodingMode.Auto()
    elif decoding_mode == "top_k":
        return trtllm.DecodingMode.TopK()
    elif decoding_mode == "top_p":
        return trtllm.DecodingMode.TopP()
    elif decoding_mode == "top_k_top_p":
        return trtllm.DecodingMode.TopKTopP()
    elif decoding_mode == "beam_search":
        return trtllm.DecodingMode.BeamSearch()
    elif decoding_mode == "medusa":
        return trtllm.DecodingMode.Medusa()
    elif decoding_mode == "redrafter":
        return trtllm.DecodingMode.ExplicitDraftTokens()
    elif decoding_mode == "lookahead":
        return trtllm.DecodingMode.Lookahead()
    elif decoding_mode == "eagle":
        return trtllm.DecodingMode.Eagle()
    raise pb_utils.TritonModelException(
        f"decoding_mode value of '{decoding_mode}' is not supported."
    )


def convert_timestamp_to_seconds(timestamp: str):
    return int(datetime.datetime.strptime(timestamp, "%m-%d-%Y %H:%M:%S.%f").timestamp())


def triton_string_to_torch(dtype):
    type_map = {
        "TYPE_BOOL": torch.bool,
        "TYPE_UINT8": torch.uint8,
        "TYPE_INT8": torch.int8,
        "TYPE_INT16": torch.int16,
        "TYPE_INT32": torch.int32,
        "TYPE_INT64": torch.int64,
        "TYPE_FP16": torch.float16,
        "TYPE_FP32": torch.float32,
        "TYPE_FP64": torch.float64,
        "TYPE_BF16": torch.bfloat16,
    }
    return type_map[dtype]


class TritonPythonModel:
    """Your Python model must use the same class name. Every Python model
    that is created must have "TritonPythonModel" as the class name.
    """

    def get_scheduler_config(self, model_config):
        batch_scheduler_policy = get_parameter(model_config, "batch_scheduler_policy")
        if batch_scheduler_policy is None:
            return trtllm.SchedulerConfig()
        return trtllm.SchedulerConfig(convert_scheduler_policy(batch_scheduler_policy))

    def get_kv_cache_config(self, model_config):
        kwargs = {
            "enable_block_reuse": get_parameter(model_config, "enable_kv_cache_reuse", bool),
            "max_tokens": get_parameter(model_config, "max_tokens_in_paged_kv_cache", int),
            "sink_token_length": get_parameter(model_config, "sink_token_length", int),
            "free_gpu_memory_fraction": get_parameter(
                model_config, "kv_cache_free_gpu_mem_fraction", float
            ),
            "cross_kv_cache_fraction": get_parameter(
                model_config, "cross_kv_cache_fraction", float
            ),
            "host_cache_size": get_parameter(model_config, "kv_cache_host_memory_bytes", int),
            "onboard_blocks": get_parameter(model_config, "kv_cache_onboard_blocks", bool),
        }
        max_attention_window_size = get_parameter(model_config, "max_attention_window_size")
        if max_attention_window_size:
            kwargs["max_attention_window"] = [int(x) for x in max_attention_window_size.split(",")]
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        return trtllm.KvCacheConfig(**kwargs)

    def get_parallel_config(self, model_config):
        kwargs = {}
        gpu_device_ids = get_parameter(model_config, "gpu_device_ids")
        if gpu_device_ids:
            kwargs["device_ids"] = [int(x) for x in gpu_device_ids.split(",")]
        self.use_orchestrator_mode = os.environ.get("TRTLLM_ORCHESTRATOR", "0") == "1"
        if self.use_orchestrator_mode:
            kwargs["communication_mode"] = trtllm.CommunicationMode.ORCHESTRATOR
            worker_path = get_parameter(model_config, "worker_path")
            spawn_processes = os.environ.get("TRTLLM_ORCHESTRATOR_SPAWN_PROCESSES", "1") == "1"
            if not spawn_processes:
                raise pb_utils.TritonModelException(
                    "Orchestrator mode with --disable-spawn-processes is not supported in the Python backend."
                )
            is_orchestrator = (mpi_rank() == 0) if spawn_processes else True
            if worker_path is not None:
                raise pb_utils.TritonModelException(
                    "worker_path parameter is specified, but this is no longer supported. Please specify executor_worker_path instead to specify the location of the trtllmExecutorWorker executable."
                )
            executor_worker_path = get_parameter(model_config, "executor_worker_path")
            kwargs["orchestrator_config"] = trtllm.OrchestratorConfig(
                is_orchestrator, executor_worker_path
            )
        if len(kwargs) > 0:
            return trtllm.ParallelConfig(**kwargs)
        return None

    def get_peft_cache_config(self, model_config):
        kwargs = {
            "optimal_adapter_size": get_parameter(
                model_config, "lora_cache_optimal_adapter_size", int
            ),
            "max_adapter_size": get_parameter(model_config, "lora_cache_max_adapter_size", int),
            "device_cache_percent": get_parameter(
                model_config, "lora_cache_gpu_memory_fraction", float
            ),
            "host_cache_size": get_parameter(model_config, "lora_cache_host_memory_bytes", int),
            "lora_prefetch_dir": get_parameter(model_config, "lora_prefetch_dir", int),
        }
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        return trtllm.PeftCacheConfig(**kwargs)

    def get_executor_lookahead_config(self, model_config):
        lookahead_window_size = get_parameter(model_config, "lookahead_window_size", int)
        lookahead_ngram_size = get_parameter(model_config, "lookahead_ngram_size", int)
        lookahead_verification_set_size = get_parameter(
            model_config, "lookahead_verification_set_size", int
        )
        # executor_lookahead_config is not set
        if all(
            item is None
            for item in [
                lookahead_window_size,
                lookahead_ngram_size,
                lookahead_verification_set_size,
            ]
        ):
            return None

        incomplete_config = None in [
            lookahead_window_size,
            lookahead_ngram_size,
            lookahead_verification_set_size,
        ]

        assert not incomplete_config, "Please set executor_lookahead_window_size, executor_lookahead_ngram_size and executor_lookahead_verification_set_size together."

        return trtllm.LookaheadDecodingConfig(
            lookahead_window_size, lookahead_ngram_size, lookahead_verification_set_size
        )

    def get_decoding_config(self, model_config):
        decoding_mode = convert_decoding_mode(get_parameter(model_config, "decoding_mode"))
        self.executor_lookahead_config = None
        if decoding_mode == trtllm.DecodingMode.Lookahead():
            # Add LAD config
            self.executor_lookahead_config = self.get_executor_lookahead_config(model_config)
        eagle_choices = parse_eagle_choices(get_parameter(model_config, "eagle_choices"))
        kwargs = {
            "medusa_choices": parse_medusa_choices(get_parameter(model_config, "medusa_choices")),
            "eagle_config": None if eagle_choices is None else trtllm.EagleConfig(eagle_choices),
            "lookahead_decoding_config": self.executor_lookahead_config,
            "decoding_mode": decoding_mode,
        }
        print(kwargs)
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        return trtllm.DecodingConfig(**kwargs)

    def get_extended_runtime_perf_knob_config(self, model_config):
        kwargs = {
            "multi_block_mode": get_parameter(model_config, "multi_block_mode", bool),
            "enable_context_fmha_fp32_acc": get_parameter(
                model_config, "enable_context_fmha_fp32_acc", bool
            ),
            "cuda_graph_mode": get_parameter(model_config, "cuda_graph_mode", bool),
            "cuda_graph_cache_size": get_parameter(model_config, "cuda_graph_cache_size", int),
        }
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        return trtllm.ExtendedRuntimePerfKnobConfig(**kwargs)

    def get_guided_decoding_config(self, model_config):
        guided_decoding_backend = get_parameter(model_config, "guided_decoding_backend", str)

        tokenizer_dir = get_parameter(model_config, "tokenizer_dir", str)
        if guided_decoding_backend not in ["xgrammar"]:
            if tokenizer_dir:
                pb_utils.Logger.log_warn(
                    f"Guided decoding backend has not been set but tokenizer_dir is given. Tokenizer_dir will be ignored."
                )
            return None

        if guided_decoding_backend == "xgrammar":
            guided_decoding_backend = trtllm.GuidedDecodingConfig.GuidedDecodingBackend.XGRAMMAR

        if not tokenizer_dir:
            raise ValueError(
                "Guided decoding requires tokenizer's information. Please provide 'tokenizer_dir'."
            )
        from transformers import AutoTokenizer

        tokenizer = AutoTokenizer.from_pretrained(tokenizer_dir)
        pb_utils.Logger.log_info(
            f"Guided decoding has been set with {guided_decoding_backend} backend"
        )
        return trtllm.GuidedDecodingConfig(
            backend=guided_decoding_backend, **_xgrammar_tokenizer_info(tokenizer)
        )

    def get_executor_config(self, model_config):
        kwargs = {
            "max_beam_width": get_parameter(model_config, "max_beam_width", int),
            "scheduler_config": self.get_scheduler_config(model_config),
            "kv_cache_config": self.get_kv_cache_config(model_config),
            "enable_chunked_context": get_parameter(model_config, "enable_chunked_context", bool),
            "normalize_log_probs": get_parameter(model_config, "normalize_log_probs", bool),
            "batching_type": convert_batching_type(get_parameter(model_config, "gpt_model_type")),
            "parallel_config": self.get_parallel_config(model_config),
            "peft_cache_config": self.get_peft_cache_config(model_config),
            "decoding_config": self.get_decoding_config(model_config),
            "max_queue_size": model_config.get(
                "dynamic_batching",
                {},
            )
            .get(
                "default_queue_policy",
                {},
            )
            .get("max_queue_size"),
            "extended_runtime_perf_knob_config": self.get_extended_runtime_perf_knob_config(
                model_config
            ),
            "guided_decoding_config": self.get_guided_decoding_config(model_config),
        }
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        return trtllm.ExecutorConfig(**kwargs)

    def create_metrics(self, model: str, version: str, is_v1_model: bool):
        self.request_metric_family = pb_utils.MetricFamily(
            name="nv_trt_llm_request_metrics",
            description="TRT LLM request metrics",
            kind=pb_utils.MetricFamily.GAUGE,
        )
        self.runtime_memory_metric_family = pb_utils.MetricFamily(
            name="nv_trt_llm_runtime_memory_metrics",
            description="TRT LLM runtime memory metrics",
            kind=pb_utils.MetricFamily.GAUGE,
        )
        self.kv_cache_metric_family = pb_utils.MetricFamily(
            name="nv_trt_llm_kv_cache_block_metrics",
            description="TRT LLM KV cache block metrics",
            kind=pb_utils.MetricFamily.GAUGE,
        )
        model_type = "v1" if is_v1_model else "inflight_batcher"
        self.model_type_metric_family = pb_utils.MetricFamily(
            name=f"nv_trt_llm_{model_type}_metrics",
            description=f"TRT LLM {model_type}-specific metrics",
            kind=pb_utils.MetricFamily.GAUGE,
        )
        self.general_metric_family = pb_utils.MetricFamily(
            name="nv_trt_llm_general_metrics",
            description="General TRT LLM metrics",
            kind=pb_utils.MetricFamily.GAUGE,
        )
        # Set the metric using self.general_metric_output_family.observe(string_size)
        self.request_tokens_metric_family = pb_utils.MetricFamily(
            name="nv_llm_input_token_len",
            description="TRT LLM response metrics",
            kind=pb_utils.MetricFamily.HISTOGRAM,
        )
        self.response_tokens_metric_family = pb_utils.MetricFamily(
            name="nv_llm_output_token_len",
            description="TRT LLM response metrics",
            kind=pb_utils.MetricFamily.HISTOGRAM,
        )
        common_labels = {"model": model, "version": version}
        self.all_metrics = {
            # Request metrics
            "num_active_requests": self.request_metric_family.Metric(
                labels={"request_type": "active", **common_labels}
            ),
            "max_num_active_requests": self.request_metric_family.Metric(
                labels={"request_type": "max", **common_labels}
            ),
            "num_scheduled_requests": self.request_metric_family.Metric(
                labels={"request_type": "scheduled", **common_labels}
            ),
            "num_context_requests": self.request_metric_family.Metric(
                labels={"request_type": "context", **common_labels}
            ),
            "num_waiting_requests": self.request_metric_family.Metric(
                labels={"request_type": "waiting", **common_labels}
            ),
            # Runtime metrics
            "cpu_mem_usage": self.runtime_memory_metric_family.Metric(
                labels={"memory_type": "cpu", **common_labels}
            ),
            "gpu_mem_usage": self.runtime_memory_metric_family.Metric(
                labels={"memory_type": "gpu", **common_labels}
            ),
            "pinned_mem_usage": self.runtime_memory_metric_family.Metric(
                labels={"memory_type": "pinned", **common_labels}
            ),
            # KV cache metrics
            "max_num_blocks": self.kv_cache_metric_family.Metric(
                labels={"kv_cache_block_type": "max", **common_labels}
            ),
            "free_num_blocks": self.kv_cache_metric_family.Metric(
                labels={"kv_cache_block_type": "free", **common_labels}
            ),
            "used_num_blocks": self.kv_cache_metric_family.Metric(
                labels={"kv_cache_block_type": "used", **common_labels}
            ),
            "tokens_per_block": self.kv_cache_metric_family.Metric(
                labels={"kv_cache_block_type": "tokens_per", **common_labels}
            ),
            "fraction_used_blocks": self.kv_cache_metric_family.Metric(
                labels={"kv_cache_block_type": "fraction", **common_labels}
            ),
            # General metrics
            "timestamp": self.general_metric_family.Metric(
                labels={"general_type": "timestamp", **common_labels}
            ),
            "iter": self.general_metric_family.Metric(
                labels={"general_type": "iteration_counter", **common_labels}
            ),
            METRIC_TOTAL_OUTPUT_TOKENS: self.response_tokens_metric_family.Metric(
                labels={"response_metric_type": METRIC_TOTAL_OUTPUT_TOKENS, **common_labels},
                buckets=build_1_2_5_buckets(1000),
            ),
            METRIC_TOTAL_INPUT_TOKENS: self.request_tokens_metric_family.Metric(
                labels={"response_metric_type": METRIC_TOTAL_INPUT_TOKENS, **common_labels},
                buckets=build_1_2_5_buckets(1000),
            ),
        }
        if is_v1_model:
            self.all_metrics.update(
                {
                    "num_ctx_tokens": self.model_type_metric_family.Metric(
                        labels={"v1_specific_metric": "total_context_tokens", **common_labels}
                    ),
                    "num_gen_tokens": self.model_type_metric_family.Metric(
                        labels={"v1_specific_metric": "total_generation_tokens", **common_labels}
                    ),
                    "empty_gen_slots": self.model_type_metric_family.Metric(
                        labels={"v1_specific_metric": "empty_generation_slots", **common_labels}
                    ),
                }
            )
        else:
            self.all_metrics.update(
                {
                    "num_ctx_tokens": self.model_type_metric_family.Metric(
                        labels={
                            "inflight_batcher_specific_metric": "total_context_tokens",
                            **common_labels,
                        }
                    ),
                    "num_gen_requests": self.model_type_metric_family.Metric(
                        labels={
                            "inflight_batcher_specific_metric": "generation_requests",
                            **common_labels,
                        }
                    ),
                    "micro_batch_id": self.model_type_metric_family.Metric(
                        labels={
                            "inflight_batcher_specific_metric": "micro_batch_id",
                            **common_labels,
                        }
                    ),
                    "num_paused_requests": self.model_type_metric_family.Metric(
                        labels={
                            "inflight_batcher_specific_metric": "paused_requests",
                            **common_labels,
                        }
                    ),
                }
            )

    def initialize(self, args):
        """`initialize` is called only once when the model is being loaded.
        Implementing `initialize` function is optional. This function allows
        the model to initialize any state associated with this model.

        Parameters
        ----------
        args : dict
          Both keys and values are strings. The dictionary keys and values are:
          * model_config: A JSON string containing the model configuration
          * model_instance_kind: A string containing model instance kind
          * model_instance_device_id: A string containing model instance device ID
          * model_repository: Model repository path
          * model_version: Model version
          * model_name: Model name
        """
        model_config = json.loads(args["model_config"])
        logger.info(f"model_config: {model_config}")
        gpt_model_path = get_parameter(model_config, "gpt_model_path")
        if get_parameter(model_config, "enable_trt_overlap", bool):
            raise pb_utils.TritonModelException(f"enable_trt_overlap=true is not supported.")
        self.exclude_input_from_output = get_parameter(
            model_config, "exclude_input_in_output", bool
        )
        executor_config = self.get_executor_config(model_config)
        self.executor = trtllm.Executor(
            gpt_model_path, trtllm.ModelType.DECODER_ONLY, executor_config
        )
        self.decoupled = pb_utils.using_decoupled_model_transaction_policy(model_config)
        logger.info(f"decoupled:{self.decoupled}")
        self.cancellation_check_period_ms = (
            get_parameter(model_config, "cancellation_check_period_ms", int) or 100
        )
        self.stats_check_period_ms = (
            get_parameter(model_config, "stats_check_period_ms", int) or 100
        )

        self.logits_dtype = None
        for output in model_config["output"]:
            if output["name"] == "context_logits" or output["name"] == "generation_logits":
                self.logits_dtype = triton_string_to_torch(output["data_type"])

        self.create_metrics(
            args["model_name"],
            args["model_version"],
            is_v1_model=executor_config.batching_type == trtllm.BatchingType.STATIC,
        )
        self.triton_user_id_to_req_ids = {}
        self.triton_req_id_to_req_ids = {}
        self.req_id_to_request_data = {}
        self.lock = Lock()
        self.running = False
        self.awaiter_thread = Thread(target=self.awaiter_loop)
        self.cancellation_thread = Thread(target=self.cancellation_loop)
        self.metrics_thread = Thread(target=self.metrics_loop)
        if self.executor.can_enqueue_requests():
            self.running = True
            self.awaiter_thread.start()
            self.cancellation_thread.start()
            self.metrics_thread.start()
        else:
            # In leader mode, worker ranks will wait here until leader is done.
            self.executor.shutdown()

    def handle_stop_request(self, triton_user_id, response_sender):
        if triton_user_id is None or triton_user_id == "":
            response_sender.send(
                pb_utils.InferenceResponse(
                    error=pb_utils.TritonError(
                        "A request id must be provided for request cancellation"
                    )
                ),
                flags=pb_utils.TRITONSERVER_RESPONSE_COMPLETE_FINAL,
            )
            return

        with self.lock:
            if triton_user_id in self.triton_user_id_to_req_ids:
                req_ids = self.triton_user_id_to_req_ids[triton_user_id]
                for req_id in req_ids:
                    self.executor.cancel_request(req_id)

        response_sender.send(
            pb_utils.InferenceResponse(), flags=pb_utils.TRITONSERVER_RESPONSE_COMPLETE_FINAL
        )

    def execute(self, requests):
        """`execute` must be implemented in every Python model. `execute`
        function receives a list of pb_utils.InferenceRequest as the only
        argument. This function is called when an inference is requested
        for this model.

        Parameters
        ----------
        requests : list
          A list of pb_utils.InferenceRequest

        Returns
        -------
        list
          A list of pb_utils.InferenceResponse. The length of this list must
          be the same as `requests`
        """
        if not self.executor.can_enqueue_requests():
            return

        # Convert to executor requests.

        triton_requests = []
        executor_requests = []
        batch_indices = []
        triton_user_ids = []
        triton_req_ids = []

        for request in requests:
            triton_user_id = request.request_id()

            response_sender = request.get_response_sender()
            stop = get_input_scalar_by_name(request, "stop")

            if stop:
                self.handle_stop_request(triton_user_id, response_sender)
            else:
                # Unique request id used to identify each triton request
                triton_req_id = str(randint(0, sys.maxsize))
                self.triton_req_id_to_req_ids[triton_req_id] = set()
                if triton_user_id is not None and triton_user_id != "":
                    self.triton_user_id_to_req_ids[triton_user_id] = set()

                try:
                    converted_reqs = convert_request(
                        request,
                        self.exclude_input_from_output,
                        self.decoupled,
                        self.executor_lookahead_config,
                    )
                except Exception as e:
                    error_message = traceback.format_exc()
                    response_sender.send(
                        pb_utils.InferenceResponse(
                            error=pb_utils.TritonError(
                                f"An error occurred when processing the input values for request id {request.request_id()}, error: {e} trace: {error_message}"
                            )
                        ),
                        flags=pb_utils.TRITONSERVER_RESPONSE_COMPLETE_FINAL,
                    )
                else:
                    for batch_index, converted_req in enumerate(converted_reqs):
                        triton_requests.append(request)
                        executor_requests.append(converted_req)
                        triton_user_ids.append(triton_user_id)
                        triton_req_ids.append(triton_req_id)
                        batch_indices.append(batch_index)

        with self.lock:
            request_ids = self.executor.enqueue_requests(executor_requests)
            # print(f"enqueue request_ids:{request_ids}")
            for (
                req_id,
                triton_req_id,
                triton_user_id,
                executor_request,
                triton_request,
                batch_index,
            ) in zip(
                request_ids,
                triton_req_ids,
                triton_user_ids,
                executor_requests,
                triton_requests,
                batch_indices,
            ):
                self.req_id_to_request_data[req_id] = RequestData(
                    triton_req_id,
                    triton_user_id,
                    batch_index,
                    len(batch_indices),
                    executor_request.sampling_config.num_return_sequences,
                    0,
                    0,
                    triton_request.get_response_sender(),
                )
                self.triton_req_id_to_req_ids[triton_req_id].add(req_id)
                input_len = (
                    len(executor_request.input_token_ids)
                    if executor_request.input_token_ids is not None
                    else 0
                )
                self.req_id_to_request_data[req_id].num_input_tokens += input_len
                # This checks both request level and instance config level
                if (
                    executor_request.output_config.exclude_input_from_output is False
                    and executor_request.streaming is False
                ):
                    self.req_id_to_request_data[req_id].num_output_tokens -= (
                        self.req_id_to_request_data[req_id].num_input_tokens
                        * executor_request.sampling_config.beam_width
                    )
                if triton_user_id is not None and triton_user_id != "":
                    self.triton_user_id_to_req_ids[triton_user_id].add(req_id)

        return None

    def awaiter_loop(self):
        """Gets responses from executor and returns the results."""
        while self.running:
            for response in self.executor.await_responses(
                timeout=datetime.timedelta(milliseconds=1)
            ):
                req_id = response.request_id
                request_data = None
                with self.lock:
                    if req_id not in self.req_id_to_request_data:
                        continue
                    request_data = self.req_id_to_request_data[req_id]

                triton_response, is_final, output_length = convert_response(
                    response,
                    request_data.batch_index,
                    request_data.batch_size,
                    request_data.num_return_sequences,
                    self.logits_dtype,
                )
                with self.lock:
                    self.req_id_to_request_data[req_id].num_output_tokens += output_length
                triton_request_final = False
                if is_final:
                    with self.lock:
                        # Check if all executor requests part of that triton request are finished
                        self.triton_req_id_to_req_ids[request_data.triton_req_id].remove(req_id)
                        if len(self.triton_req_id_to_req_ids[request_data.triton_req_id]) == 0:
                            pb_utils.Logger.log_info(
                                f"DELETING Req id {req_id}, triton_req_id {request_data.triton_req_id} "
                            )
                            triton_request_final = True
                            del self.triton_req_id_to_req_ids[request_data.triton_req_id]
                            if (
                                request_data.triton_user_id is not None
                                and request_data.triton_user_id != ""
                            ):
                                del self.triton_user_id_to_req_ids[request_data.triton_user_id]
                        self.update_metrics_per_request(req_id)
                        del self.req_id_to_request_data[req_id]
                # print(f"triton_response:{triton_response}")

                request_data.response_sender.send(
                    triton_response,
                    flags=pb_utils.TRITONSERVER_RESPONSE_COMPLETE_FINAL
                    if triton_request_final
                    else 0,
                )

    def cancellation_loop(self):
        """Checks if any pending requests have been cancelled."""
        while self.running:
            time.sleep(self.cancellation_check_period_ms / 1000.0)
            with self.lock:
                for req_id, request_data in self.req_id_to_request_data.items():
                    if request_data.response_sender.is_cancelled():
                        self.executor.cancel_request(req_id)

    def update_metrics_per_request(self, req_id):
        """Updates triton metrics after completing one request"""
        output_tokens = self.req_id_to_request_data[req_id].num_output_tokens
        input_tokens = self.req_id_to_request_data[req_id].num_input_tokens

        self.all_metrics[METRIC_TOTAL_OUTPUT_TOKENS].observe(output_tokens)
        self.all_metrics[METRIC_TOTAL_INPUT_TOKENS].observe(input_tokens)

    def get_composite_metric_map(self, stat):
        def get_metric(metric_name, family_stats=None):
            if family_stats is None:
                if hasattr(stat, metric_name):
                    return getattr(stat, metric_name)
                elif stat.kv_cache_stats is not None and hasattr(stat.kv_cache_stats, metric_name):
                    return getattr(stat.kv_cache_stats, metric_name)
                elif stat.static_batching_stats is not None and hasattr(
                    stat.static_batching_stats, metric_name
                ):
                    return getattr(stat.static_batching_stats, metric_name)
                elif stat.inflight_batching_stats is not None and hasattr(
                    stat.inflight_batching_stats, metric_name
                ):
                    return getattr(stat.inflight_batching_stats, metric_name)
            elif family_stats is not None and hasattr(family_stats, metric_name):
                return getattr(family_stats, metric_name)
            pb_utils.Logger.log_warn(f'Constituent metric "{metric_name}" not found.')
            return None

        composite_metrics = {}

        # compute fraction_used_blocks
        max_blocks = get_metric("max_num_blocks", stat.kv_cache_stats)
        used_blocks = get_metric("used_num_blocks", stat.kv_cache_stats)
        if max_blocks is not None and used_blocks is not None:
            composite_metrics["fraction_used_blocks"] = (
                0.0 if max_blocks <= 0 else used_blocks / max_blocks
            )
        else:
            pb_utils.Logger.log_warn(
                f"fraction_used_blocks is missing one or more constituent metric."
            )

        # compute num_waiting_requests
        active_requests = get_metric("num_active_requests")
        scheduled_requests = get_metric("num_scheduled_requests")
        if active_requests is not None and scheduled_requests is not None:
            composite_metrics["num_waiting_requests"] = active_requests - scheduled_requests
        else:
            pb_utils.Logger.log_warn(
                f"num_waiting_requests is missing one or more constituent metric."
            )

        return composite_metrics

    def metrics_loop(self):
        """Updates triton metrics using stats from the executor."""
        while self.running:
            time.sleep(self.stats_check_period_ms / 1000.0)
            for stat in self.executor.get_latest_iteration_stats():
                try:
                    composite_metrics = self.get_composite_metric_map(stat)
                    for key, metric in self.all_metrics.items():
                        # Skip processing for both histogram metrics
                        if isinstance(key, str) and key in [
                            METRIC_TOTAL_OUTPUT_TOKENS,
                            METRIC_TOTAL_INPUT_TOKENS,
                        ]:
                            continue
                        value = None
                        if hasattr(stat, key):
                            value = getattr(stat, key)
                        elif stat.kv_cache_stats is not None and hasattr(stat.kv_cache_stats, key):
                            value = getattr(stat.kv_cache_stats, key)
                        elif stat.static_batching_stats is not None and hasattr(
                            stat.static_batching_stats, key
                        ):
                            value = getattr(stat.static_batching_stats, key)
                        elif stat.inflight_batching_stats is not None and hasattr(
                            stat.inflight_batching_stats, key
                        ):
                            value = getattr(stat.inflight_batching_stats, key)
                        elif key in composite_metrics:
                            value = composite_metrics[key]
                        if value is not None:
                            if key == "timestamp":
                                value = convert_timestamp_to_seconds(value)
                            metric.set(value)
                        else:
                            pb_utils.Logger.log_warn(f'Metric "{key}" not found.')
                except Exception as e:
                    pb_utils.Logger.log_warn(f"Error while processing metrics: {e}")

    def finalize(self):
        """`finalize` is called only once when the model is being unloaded.
        Implementing `finalize` function is optional. This function allows
        the model to perform any necessary clean ups before exit.
        """
        if self.executor.can_enqueue_requests():
            self.running = False
            self.awaiter_thread.join()
            self.cancellation_thread.join()
            self.metrics_thread.join()
            self.executor.shutdown()



from typing import List, Literal, Optional

from pydantic import BaseModel, Field
from google.genai import types


class LiveMultiModalInputMessage(BaseModel):
    context_id: Optional[str] = Field(default=None, description="The context ID for the message.")
    message_id: Optional[str] = Field(default=None, description="message ID")
    task_id: Optional[str] = Field(default=None, description="task ID for the message")
    kind: Literal["text", "audio", "images", "text_images", "audio_images"] = Field(
        default="audio_images", description="the task kind of the input message"
    )
    text_content: Optional[types.Content] = Field(
        default=None,
        description="""Optional. text content when kind is text""",
    )
    audio_blob: Optional[types.Blob] = Field(
        default=None,
        description="""Optional. audio binary blob when kind is audio or audio_images""",
    )
    image_blob_list: Optional[List[types.Blob]] = Field(
        default=None,
        description="""Optional. image binary blob list when kind is text_images or audio_images""",
    )


class LiveMultiModalOutputMessage(BaseModel):
    context_id: Optional[str] = Field(default=None, description="The context ID for the message.")
    message_id: Optional[str] = Field(default=None, description="message ID")
    task_id: Optional[str] = Field(default=None, description="task ID for the message")
    kind: Literal["transcription", "text", "audio", "images", "interrupted"] = Field(
        default="text", description="the task kind of gen output"
    )
    is_first_text: bool = False
    text_content: Optional[types.Content] = Field(
        default=None,
        description="""Optional. text content when kind is transcription or text""",
    )
    is_first_audio_chunk: bool = False
    audio_blob: Optional[types.Blob] = Field(
        default=None,
        description="""Optional. audio binary blob when kind is audio""",
    )
    image_blob_list: Optional[List[types.Blob]] = Field(
        default=None,  # NOTE: gemini2.5 live don't gen image, mayge feature gen
        description="""Optional. image binary blob list when kind is images""",
    )
    interrupted: Optional[bool] = Field(
        default=None,
        description="""Flag indicating that LLM was interrupted when generating the content.
    Usually it's due to user interruption during a bidi streaming.""",
    )



import logging

from apipeline.pipeline.pipeline import Pipeline
from apipeline.pipeline.task import PipelineParams, PipelineTask
from apipeline.pipeline.runner import PipelineRunner
from fastapi import WebSocket

from achatbot.cmd.bots.base_fastapi_websocket_server import AIFastapiWebsocketBot
from achatbot.processors.aggregators.llm_response import (
    LLMAssistantResponseAggregator,
    LLMUserResponseAggregator,
)
from achatbot.processors.llm.base import LLMProcessor
from achatbot.processors.speech.tts.tts_processor import TTSProcessor
from achatbot.modules.speech.vad_analyzer import VADAnalyzerEnvInit
from achatbot.types.frames.data_frames import LLMMessagesFrame
from achatbot.cmd.bots import register_ai_fastapi_ws_bots

from dotenv import load_dotenv

from achatbot.types.network.fastapi_websocket import FastapiWebsocketServerParams
from achatbot.transports.fastapi_websocket_server import FastapiWebsocketTransport

load_dotenv(override=True)


@register_ai_fastapi_ws_bots.register
class FastapiWebsocketServerBot(AIFastapiWebsocketBot):
    """
    fastapi websocket input/output server bot with vad,asr,llm,tts
    """

    def __init__(self, websocket: WebSocket | None = None, **args) -> None:
        super().__init__(websocket=websocket, **args)
        self.init_bot_config()

        self.vad_analyzer = self.get_vad_analyzer()
        self.asr_processor = self.get_asr_processor()
        self.llm_processor: LLMProcessor = self.get_llm_processor()
        self.tts_processor: TTSProcessor = self.get_tts_processor()

    async def arun(self):
        if self._websocket is None:
            return

        self.params = FastapiWebsocketServerParams(
            audio_in_enabled=True,
            audio_out_enabled=True,
            add_wav_header=True,
            vad_enabled=True,
            vad_analyzer=self.vad_analyzer,
            vad_audio_passthrough=True,
        )
        stream_info = self.tts_processor.get_stream_info()
        self.params.audio_out_sample_rate = stream_info["sample_rate"]
        self.params.audio_out_channels = stream_info["channels"]
        transport = FastapiWebsocketTransport(
            websocket=self._websocket,
            params=self.params,
        )

        messages = []
        if self._bot_config.llm.messages:
            messages = self._bot_config.llm.messages
        user_response = LLMUserResponseAggregator(messages)
        assistant_response = LLMAssistantResponseAggregator(messages)

        self.task = PipelineTask(
            Pipeline(
                [
                    transport.input_processor(),
                    self.asr_processor,
                    user_response,
                    self.llm_processor,
                    self.tts_processor,
                    transport.output_processor(),
                    assistant_response,
                ]
            ),
            params=PipelineParams(
                allow_interruptions=True,
                enable_metrics=True,
                send_initial_empty_metrics=False,
            ),
        )

        transport.add_event_handler("on_client_connected", self.on_client_connected)
        transport.add_event_handler("on_client_disconnected", self.on_client_disconnected)

        await PipelineRunner(handle_sigint=self._handle_sigint).run(self.task)

    async def on_client_connected(
        self,
        transport: FastapiWebsocketTransport,
        websocket: WebSocket,
    ):
        logging.info(f"on_client_connected client:{websocket.client}")
        self.session.set_client_id(client_id=f"{websocket.client.host}:{websocket.client.port}")

        # joined use tts say "hello" to introduce with llm generate
        if self._bot_config.tts and self._bot_config.llm and self._bot_config.llm.messages:
            hi_text = "Please introduce yourself first."
            if self._bot_config.llm.language and self._bot_config.llm.language == "zh":
                hi_text = "请用中文介绍下自己。"
            self._bot_config.llm.messages.append(
                {
                    "role": "user",
                    "content": hi_text,
                }
            )
            await self.task.queue_frames([LLMMessagesFrame(self._bot_config.llm.messages)])



import argparse
import asyncio
import queue
import sys
import signal

import sphn
import aiohttp
import sounddevice as sd
import numpy as np

from apipeline.frames.data_frames import AudioRawFrame, TextFrame
from apipeline.serializers.protobuf import ProtobufFrameSerializer

# Global flag for shutdown
shutdown_flag = asyncio.Event()

# ANSI escape codes for colored text
GREEN = "\033[92m"
RESET = "\033[0m"


class Connection:
    """
    Connection manager
    """

    def __init__(self, ws: aiohttp.ClientWebSocketResponse):
        self.ws = ws
        self.sample_rate = 24000
        self.frame_size = 1920
        self.channels = 1

        # The Opus audio codec is used for streaming audio to the websocket
        # For spoken voice, it can compress as much as 10x, and can be decoded in real-time
        self.opus_writer = sphn.OpusStreamWriter(self.sample_rate)
        self.opus_reader = sphn.OpusStreamReader(self.sample_rate)

        self.audio_in_stream = sd.InputStream(
            samplerate=self.sample_rate,
            channels=self.channels,
            blocksize=self.frame_size,
            callback=self.audio_in_callback,
        )

        self.audio_out_stream = sd.OutputStream(
            samplerate=self.sample_rate,
            channels=self.channels,
            blocksize=self.frame_size,
            callback=self.audio_out_callback,
        )
        self.out_queue = queue.Queue()

        self.serializer = ProtobufFrameSerializer()

    # Sounddevice callbacks for handling raw audio to/from the speaker and mic
    def audio_in_callback(self, data, frames, time, status):
        self.opus_writer.append_pcm(data[:, 0])

    def audio_out_callback(self, data, frames, time, status):
        try:
            pcm_data = self.out_queue.get(block=False)
            assert pcm_data.shape == (self.frame_size,), pcm_data.shape
            data[:, 0] = pcm_data
        except queue.Empty:
            data.fill(0)

    # Async loops for bidirectional audio streaming:

    async def send_loop(self):
        """
        Async loop for sending opus stream to the websocket
        """
        while not shutdown_flag.is_set():
            await asyncio.sleep(0.001)
            msg = self.opus_writer.read_bytes()
            if len(msg) > 0:
                try:
                    # send audio frame
                    audio_frame = AudioRawFrame(
                        audio=msg,
                        sample_rate=self.sample_rate,
                        num_channels=self.channels,
                    )
                    bytes_audio_frame = self.serializer.serialize(audio_frame)
                    # print("send serialize __audio_frame__", bytes_audio_frame)
                    await self.ws.send_bytes(bytes_audio_frame)
                except Exception as e:
                    print(f"Error in send_loop: {e}")
                    break
        print("send_loop done")
        if not self.ws.closed:
            await self.ws.close()

    async def receive_loop(self):
        """
        Async loop for receiving messages from the websocket, including text and opus stream
        """
        sentence = ""
        async for msg in self.ws:
            try:
                if shutdown_flag.is_set():
                    break
                msg_bytes = msg.data
                if not isinstance(msg_bytes, bytes) or len(msg_bytes) == 0:
                    continue

                # receive frame
                frame = self.serializer.deserialize(msg_bytes)
                # print("____deserialize receive frame____", frame)
                if isinstance(frame, AudioRawFrame):
                    # payload is opus audio
                    self.opus_reader.append_bytes(frame.audio)

                if isinstance(frame, TextFrame):
                    # payload is text output from the model, print it to the console
                    token = frame.text
                    sentence += token
                    sys.stdout.write(f"\r{GREEN}{sentence.lstrip()}{RESET}")
                    sys.stdout.flush()
                    if sentence.strip()[-1] in [".", "!", "?"]:
                        sys.stdout.write("\n")
                        sentence = ""
            except Exception as e:
                print(f"Error in receive_loop: {e}")
                break

        print("receive_loop done")

    async def decoder_loop(self):
        """
        Async loop for decoding audio from the websocket into raw pcm audio, and queueing it for playback
        """

        all_pcm_data = None
        while not shutdown_flag.is_set():
            await asyncio.sleep(0.001)
            pcm = self.opus_reader.read_pcm()
            if pcm is None:
                continue

            if all_pcm_data is None:
                all_pcm_data = pcm
            else:
                all_pcm_data = np.concatenate((all_pcm_data, pcm))
            while all_pcm_data.shape[-1] >= self.frame_size:
                self.out_queue.put(all_pcm_data[: self.frame_size])
                all_pcm_data = np.array(all_pcm_data[self.frame_size :])
        print("decoder_loop done")

    async def run(self):
        try:
            with self.audio_in_stream, self.audio_out_stream:
                self.futures = asyncio.gather(
                    self.send_loop(), self.receive_loop(), self.decoder_loop()
                )
                await self.futures
        except asyncio.CancelledError:
            print("Connection tasks cancelled")


def sigint_handler(signum, frame):
    # Handle keyboard interrupts
    print("\nEnding conversation...")
    shutdown_flag.set()


async def run(endpoint: str):
    signal.signal(signal.SIGINT, sigint_handler)

    print("Connecting to", endpoint)
    print("This may trigger a cold boot of the model...\n")
    async with aiohttp.ClientSession() as session:
        try:
            async with session.ws_connect(endpoint) as ws:
                connection = Connection(ws)
                print("Connection established.")
                print("Conversation started. Press Ctrl+C to exit.\n")
                await connection.run()
        except aiohttp.ClientError as e:
            print(f"Connection error: {e}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Moshi Websocket Client")
    parser.add_argument(
        "--endpoint", type=str, default="ws://localhost:4321", help="websocket endpoint"
    )
    args = parser.parse_args()

    try:
        asyncio.run(run(args.endpoint))
    except KeyboardInterrupt:
        print("\nProgram interrupted")
    finally:
        print("Conversation complete")



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from dataclasses import dataclass, field
from typing import Optional, List
from omegaconf import II

from fairseq.dataclass import FairseqDataclass
from fairseq.optim.lr_scheduler import FairseqLRScheduler, register_lr_scheduler


@dataclass
class FixedLRScheduleConfig(FairseqDataclass):
    force_anneal: Optional[int] = field(
        default=None,
        metadata={"help": "force annealing at specified epoch"},
    )
    lr_shrink: float = field(
        default=0.1,
        metadata={"help": "shrink factor for annealing, lr_new = (lr * lr_shrink)"},
    )
    warmup_updates: int = field(
        default=0,
        metadata={"help": "warmup the learning rate linearly for the first N updates"},
    )
    lr: List[float] = II("optimization.lr")


@register_lr_scheduler("fixed", dataclass=FixedLRScheduleConfig)
class FixedLRSchedule(FairseqLRScheduler):
    """Decay the LR on a fixed schedule."""

    def __init__(self, cfg: FixedLRScheduleConfig, optimizer):
        super().__init__(cfg, optimizer)

        self.lr = cfg.lr[0]
        if cfg.warmup_updates > 0:
            self.warmup_factor = 1.0 / cfg.warmup_updates
        else:
            self.warmup_factor = 1

    def state_dict(self):
        return {"lr": self.lr}

    def load_state_dict(self, state_dict):
        if "lr" in state_dict:
            self.lr = state_dict["lr"]

    def get_next_lr(self, epoch):
        lrs = self.cfg.lr
        if self.cfg.force_anneal is None or epoch < self.cfg.force_anneal:
            # use fixed LR schedule
            next_lr = lrs[min(epoch - 1, len(lrs) - 1)]
        else:
            # annneal based on lr_shrink
            next_lr = lrs[-1] * self.cfg.lr_shrink ** (
                epoch + 1 - self.cfg.force_anneal
            )
        return next_lr

    def step_begin_epoch(self, epoch):
        """Update the learning rate at the beginning of the given epoch."""
        self.lr = self.get_next_lr(epoch)
        self.optimizer.set_lr(self.warmup_factor * self.lr)
        return self.optimizer.get_lr()

    def step_update(self, num_updates):
        """Update the learning rate after each update."""
        if self.cfg.warmup_updates > 0 and num_updates < self.cfg.warmup_updates:
            self.warmup_factor = (num_updates + 1) / float(self.cfg.warmup_updates)
            self.optimizer.set_lr(self.warmup_factor * self.lr)
        else:
            self.optimizer.set_lr(self.lr)
        return self.optimizer.get_lr()



import os
import uuid
import json
import logging
from pathlib import Path
import subprocess

import modal

from . import model_dir, MODEL_VOL, comfyui_out_dir, COMFYUI_OUT_VOL, clear_comfyui_output_dir


# https://github.com/black-forest-labs/flux
# https://huggingface.co/black-forest-labs/FLUX.1-schnell
# https://huggingface.co/Comfy-Org/flux1-schnell/tree/main
# modal run src/download_models.py --repo-ids "Comfy-Org/flux1-schnell" --allow-patterns "flux1-schnell-fp8.safetensors"


def link_comfyui_dir():
    # symlink the model to the right ComfyUI directory
    ckpt_path = f"{model_dir}/Comfy-Org/flux1-schnell/flux1-schnell-fp8.safetensors"
    subprocess.run(
        f"ln -s {ckpt_path} /root/comfy/ComfyUI/models/checkpoints/flux1-schnell-fp8.safetensors",
        shell=True,
        check=True,
    )

    clear_comfyui_output_dir()


def change_workflow_conf(file_path: Path, **kwargs) -> str:
    workflow_data = json.loads(file_path.read_text())

    # insert the prompt
    if kwargs.get("prompt"):
        workflow_data["6"]["inputs"]["text"] = kwargs.get("prompt")

    # give the output image a unique id per client request
    client_id = uuid.uuid4().hex
    workflow_data["9"]["inputs"]["filename_prefix"] = client_id

    # change output image size: width x height
    if kwargs.get("width"):
        workflow_data["27"]["inputs"]["width"] = kwargs.get("width")
    if kwargs.get("height"):
        workflow_data["27"]["inputs"]["height"] = kwargs.get("height")

    # ksample steps
    if kwargs.get("steps"):
        workflow_data["31"]["inputs"]["steps"] = kwargs.get("steps")

    # print(f"{workflow_data=}")

    # save this updated workflow to a new file
    new_workflow_file = f"{client_id}.json"
    json.dump(workflow_data, Path(new_workflow_file).open("w", encoding="utf-8"))

    return client_id



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from google.adk.agents import Agent
from google.adk.tools import google_search  # Import the tool
from dotenv import load_dotenv

load_dotenv()

agent = Agent(
    # A unique name for the agent.
    name="google_search_agent",
    # The Large Language Model (LLM) that agent will use.
    model=os.getenv("DEMO_AGENT_MODEL"),
    # A short description of the agent's purpose.
    description="Agent to answer questions using Google Search.",
    # Instructions to set the agent's behavior.
    instruction="Answer the question using the Google Search tool.",
    # Add google_search tool to perform grounding with Google search.
    tools=[google_search],
)



# encoding: utf-8

import editdistance
import argparse
import re


def compute_wer_txt(hypo_f, ref_f):
    f = open(hypo_f)
    hypos = f.readlines()
    f.close()
    f = open(ref_f)
    refs = f.readlines()
    f.close()
    err,cnt = 0,0
    print("Start Computing WER")
    for i in range(len(hypos)):
        hypo = hypos[i].strip().split()[:-1]
        ref = refs[i].strip().split()[:-1]

        e = editdistance.eval(ref, hypo)
        #print(ref_token,predict_token,e/float(len(ref_token)+0.001))

        err += e
        cnt += len(ref)
    print(err,cnt,float(float(err+0.0) / float(cnt+0.01)*100))


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--hypo', '--h', type=str, required=True,
                help='Input file path')
    parser.add_argument('--ref', '--r', type=str, required=True)

    args = parser.parse_args()
    compute_wer_txt(args.hypo, args.ref)



import os
import logging


from dotenv import load_dotenv

from src.common import interface
from src.common.factory import EngineClass, EngineFactory
from src.types.speech.turn_analyzer.smart_turn import (
    MODELS_DIR,
    SmartTurnArgs,
)

load_dotenv(override=True)


class TurnAnalyzerEnvInit:
    @staticmethod
    def get_smart_turn_analyzer_args() -> dict:
        args = dict(
            **SmartTurnArgs(
                model_path=os.getenv(
                    "TURN_MODEL_PATH", os.path.join(MODELS_DIR, "pipecat-ai/smart-turn-v2")
                ),
                warmup_steps=int(os.getenv("TURN_WARMUP_STEPS", "2")),
                sample_rate=int(os.getenv("TURN_SAMPLE_RATE", "16000")),
                stop_secs=float(os.getenv("TURN_STOP_SECS", "3")),
                pre_speech_ms=float(os.getenv("TURN_PRE_SPEECH_MS", "0")),
                max_duration_secs=float(os.getenv("TURN_MAX_DURATION_SECS", "8")),
            )
        )
        return args

    @staticmethod
    def getEngine(tag, **kwargs) -> interface.ITurnAnalyzer | EngineClass:
        if "v2_smart_turn_analyzer" in tag:
            from . import smart_turn

        engine = EngineFactory.get_engine_by_tag(EngineClass, tag, **kwargs)
        return engine

    @staticmethod
    def initTurnAnalyzerEngine(
        tag: str | None = None, kwargs: dict | None = None
    ) -> interface.ITurnAnalyzer | EngineClass:
        tag = tag or os.getenv("TURN_ANALYZER_TAG", "v2_smart_turn_analyzer")
        kwargs = kwargs or TurnAnalyzerEnvInit.map_config_func[tag]()
        logging.info(f"initTurnAnalyzerEngine: {tag}, {kwargs}")
        engine = TurnAnalyzerEnvInit.getEngine(tag, **kwargs)
        logging.info(f"initTurnAnalyzerEngine: {tag}, {engine}")
        return engine

    map_config_func = {
        "v2_smart_turn_analyzer": get_smart_turn_analyzer_args,
    }



import logging
from typing import Callable, Literal

from a2a.server.agent_execution import AgentExecutor, RequestContext
from a2a.server.events import EventQueue
from a2a.types import (
    TaskArtifactUpdateEvent,
    TaskState,
    TaskStatus,
    TaskStatusUpdateEvent,
)
from a2a.utils import new_agent_text_message, new_task, new_text_artifact

from .agent import Agent


class GitMcpAgentExecutor(AgentExecutor):
    """Test AgentProxy Implementation."""

    def __init__(
        self,
        mode: Literal["complete", "stream"] = "stream",
        token_stream_callback: Callable[[str], None] | None = print,
        mcp_url: str | None = "https://gitmcp.io/ai-bot-pro/achatbot",
    ):
        self.agent = Agent(
            mode=mode,
            token_stream_callback=token_stream_callback,
            mcp_url=mcp_url,
        )

    async def execute(
        self,
        context: RequestContext,
        event_queue: EventQueue,
    ) -> None:
        query = context.get_user_input()
        task = context.current_task

        if not context.message:
            raise Exception("No message provided")

        logging.info(f"{query=} {context.current_task=} {context.message=}")

        if not task:
            task = new_task(context.message)
            await event_queue.enqueue_event(task)

        async for event in self.agent.stream(query):
            if event["is_task_complete"]:
                await event_queue.enqueue_event(
                    TaskArtifactUpdateEvent(
                        append=False,
                        context_id=task.context_id,
                        task_id=task.id,
                        last_chunk=True,
                        artifact=new_text_artifact(
                            name="current_result",
                            description="Result of request to agent.",
                            text=event["content"],
                        ),
                    )
                )
                await event_queue.enqueue_event(
                    TaskStatusUpdateEvent(
                        status=TaskStatus(state=TaskState.completed),
                        final=True,
                        context_id=task.context_id,
                        task_id=task.id,
                    )
                )
            elif event["require_user_input"]:
                await event_queue.enqueue_event(
                    TaskStatusUpdateEvent(
                        status=TaskStatus(
                            state=TaskState.input_required,
                            message=new_agent_text_message(
                                event["content"],
                                task.context_id,
                                task.id,
                            ),
                        ),
                        final=True,
                        context_id=task.context_id,
                        task_id=task.id,
                    )
                )
            else:
                await event_queue.enqueue_event(
                    TaskStatusUpdateEvent(
                        append=True,
                        status=TaskStatus(
                            state=TaskState.working,
                            message=new_agent_text_message(
                                event["content"],
                                task.context_id,
                                task.id,
                            ),
                        ),
                        final=False,
                        context_id=task.context_id,
                        task_id=task.id,
                    )
                )

    async def cancel(self, context: RequestContext, event_queue: EventQueue) -> None:
        raise Exception("cancel not supported")



# author: weedge (weege007@gmail.com)

import os
import subprocess
from threading import Thread
from time import perf_counter
from typing import Optional

import modal


app = modal.App("Keye-VL")
img = (
    # https://catalog.ngc.nvidia.com/orgs/nvidia/containers/cuda/tags
    modal.Image.from_registry(
        "nvidia/cuda:12.6.1-cudnn-devel-ubuntu22.04",
        add_python="3.10",
    )
    .apt_install("git", "git-lfs", "ffmpeg", "clang", "cmake", "ninja-build")
    .pip_install(
        "torch==2.6.0",
        "torchaudio==2.6.0",
        "torchvision==0.21.0",
    )
    .pip_install("wheel", "packaging")
    # https://github.com/Dao-AILab/flash-attention/releases/tag/v2.7.4.post1
    .pip_install("flash-attn==2.7.4.post1", extra_options="--no-build-isolation")
    .run_commands(
        "pip install git+https://github.com/huggingface/transformers@17b3c96c00cd8421bff85282aec32422bdfebd31"
    )
    .pip_install("accelerate", "keye-vl-utils[decord]==1.0.0")
    .env(
        {
            "PYTORCH_CUDA_ALLOC_CONF": "expandable_segments:True",
            # "TQDM_DISABLE": "1",
            "LLM_MODEL": os.getenv("LLM_MODEL", "Kwai-Keye/Keye-VL-8B-Preview"),
            "ROUND": os.getenv("ROUND", "1"),
            "IS_OUTPUT_THINK": os.getenv("IS_OUTPUT_THINK", "1"),
        }
    )
)

HF_MODEL_DIR = "/root/.achatbot/models"
hf_model_vol = modal.Volume.from_name("models", create_if_missing=True)
ASSETS_DIR = "/root/.achatbot/assets"
assets_dir = modal.Volume.from_name("assets", create_if_missing=True)
VIDEO_OUTPUT_DIR = "/gen_video"
video_out_vol = modal.Volume.from_name("gen_video", create_if_missing=True)


with img.imports():
    import torch
    from PIL import Image
    from transformers import AutoModel, AutoTokenizer, AutoProcessor
    from keye_vl_utils import process_vision_info

    from transformers.generation.streamers import TextIteratorStreamer


@app.function(
    gpu=os.getenv("IMAGE_GPU", None),
    cpu=2.0,
    retries=1,
    image=img,
    volumes={
        HF_MODEL_DIR: hf_model_vol,
        ASSETS_DIR: assets_dir,
        VIDEO_OUTPUT_DIR: video_out_vol,
    },
    timeout=1200,  # default 300s
    scaledown_window=1200,
    max_containers=1,
)
def run(func, thinking):
    subprocess.run("nvidia-smi --version", shell=True)
    subprocess.run("nvcc --version", shell=True)
    gpu_prop = torch.cuda.get_device_properties("cuda")
    print(gpu_prop)

    func(gpu_prop, thinking)


def print_model_params(model: torch.nn.Module, extra_info=""):
    # print the number of parameters in the model
    model_million_params = sum(p.numel() for p in model.parameters()) / 1e6
    print(model)
    print(f"{extra_info} {model_million_params} M parameters")


def dump_model(gpu_prop, thinking):
    """
    vlm text model use qwen3 arch no MTP
    """
    for model_name in [
        "Kwai-Keye/Keye-VL-8B-Preview",
    ]:
        MODEL_PATH = os.path.join(HF_MODEL_DIR, model_name)
        processor = AutoProcessor.from_pretrained(
            MODEL_PATH,
            trust_remote_code=True,
        )
        model = AutoModel.from_pretrained(
            MODEL_PATH,
            torch_dtype=torch.bfloat16,
            attn_implementation="flash_attention_2" if gpu_prop.major >= 8 else None,
            trust_remote_code=True,
        ).to("cuda")

        model = model.eval()
        print(f"{model.config=}")
        print(f"{processor=}")
        print_model_params(model, f"{model_name}")

        del model
        torch.cuda.empty_cache()


@torch.inference_mode()
def predict_text(gpu_prop, thinking):
    # Load model
    model_name = os.getenv("LLM_MODEL")
    MODEL_PATH = os.path.join(HF_MODEL_DIR, model_name)
    model = AutoModel.from_pretrained(
        MODEL_PATH,
        torch_dtype=torch.bfloat16,
        attn_implementation="flash_attention_2" if gpu_prop.major >= 8 else None,
        trust_remote_code=True,
    ).to("cuda")
    model = model.eval()
    tokenizer = AutoTokenizer.from_pretrained(MODEL_PATH, use_fast=True)

    # predict with no instruct tpl, use base model | sft-it | rl-it
    text = "你叫什么名字？"
    if thinking is True:
        text += "/think"
    if thinking is False:
        text += "/no_think"
    inputs = tokenizer([text], return_tensors="pt").to(model.device)
    for key, value in inputs.items():
        print(f"{key}: {value.shape=}")
    input_ids = inputs["input_ids"]
    prompt = tokenizer.decode(input_ids[0])
    print(f"{prompt=}")

    temperature = 0.6
    generated_ids = model.generate(
        **inputs,
        do_sample=True if temperature > 0 else False,
        temperature=temperature,
        top_k=20,
        top_p=0.9,
        repetition_penalty=1.1,
        max_new_tokens=128,
    )
    print(f"{generated_ids.shape=}")
    generated_text = tokenizer.decode(
        generated_ids[0][len(input_ids[0]) :],
        # skip_special_tokens=True,
    )
    print(generated_text)


@torch.inference_mode()
def chat_text(gpu_prop, thinking):
    # Load model
    model_name = os.getenv("LLM_MODEL")
    MODEL_PATH = os.path.join(HF_MODEL_DIR, model_name)
    model = AutoModel.from_pretrained(
        MODEL_PATH,
        torch_dtype=torch.bfloat16,
        attn_implementation="flash_attention_2" if gpu_prop.major >= 8 else None,
        trust_remote_code=True,
    ).to("cuda")
    model = model.eval()
    tokenizer = AutoTokenizer.from_pretrained(MODEL_PATH, use_fast=True)

    # Construct prompt
    text = "你叫什么名字？"
    # text = "你叫什么名字？/auto_think"
    if thinking is True:
        text += "/think"
    if thinking is False:
        text += "/no_think"
    messages = [
        {
            "role": "system",
            "content": "你是一个非常棒的聊天助手，不要使用特殊字符回复。",
        },
        {
            "role": "user",
            "content": text,
        },
    ]

    # use instruct sft | rl model
    prompt = tokenizer.apply_chat_template(
        messages,
        add_generation_prompt=True,
        tokenize=False,
    )
    print(f"{prompt=}")
    inputs = tokenizer([prompt], return_tensors="pt").to(model.device)
    for key, value in inputs.items():
        print(f"{key}: {value.shape=}")
    input_ids = inputs["input_ids"]

    temperature = 0.6
    generated_ids = model.generate(
        **inputs,
        do_sample=True if temperature > 0 else False,
        temperature=temperature,
        top_k=20,
        top_p=0.9,
        repetition_penalty=1.1,
        max_new_tokens=1024,
    )
    print(f"{generated_ids.shape=}")
    generated_text = tokenizer.decode(
        generated_ids[0][len(input_ids[0]) :],
        # skip_special_tokens=True,
    )
    print(generated_text)


@torch.inference_mode()
def predict(gpu_prop, thinking):
    # Load model
    model_name = os.getenv("LLM_MODEL")
    MODEL_PATH = os.path.join(HF_MODEL_DIR, model_name)
    processor = AutoProcessor.from_pretrained(
        MODEL_PATH,
        use_fast=True,
        trust_remote_code=True,
    )
    model = AutoModel.from_pretrained(
        MODEL_PATH,
        torch_dtype=torch.bfloat16,
        _attn_implementation="flash_attention_2" if gpu_prop.major >= 8 else None,
        trust_remote_code=True,
    ).to("cuda")

    model = model.eval()
    # print(f"{model.config=}")
    # print(f"{processor=}")
    # print_model_params(model, f"{model_name}")

    # Construct prompt
    # text = "请用中文描述图片内容"
    text = f"请用中文描述图片内容，不要使用特殊字符回复。"
    if thinking is True:
        text += "/think"
    if thinking is False:
        text += "/no_think"

    # don't to chat with smolvlm, just do vision task
    # text = "Please reply to my message in Chinese simplified(简体中文), don't use Markdown format. 你好"

    messages = [
        {
            "role": "user",
            "content": [
                {
                    "type": "image",
                    "url": "https://huggingface.co/datasets/huggingface/documentation-images/resolve/main/bee.jpg",
                },
                {"type": "text", "text": text},
            ],
        },
    ]

    inputs = processor.apply_chat_template(
        messages,
        add_generation_prompt=True,
        tokenize=True,
        return_dict=True,
        return_tensors="pt",
    ).to(model.device, dtype=torch.bfloat16)
    for key, value in inputs.items():
        print(f"{key}: {value.shape=}")
    input_ids = inputs["input_ids"]
    prompt = processor.decode(input_ids[0])
    print(f"{prompt=}")

    temperature = 0.6
    generated_ids = model.generate(
        **inputs,
        do_sample=True if temperature > 0 else False,
        temperature=temperature,
        top_k=20,
        top_p=0.9,
        repetition_penalty=1.1,
        max_new_tokens=1024,
    )
    print(f"{generated_ids.shape=}")
    generated_text = processor.decode(
        generated_ids[0][len(input_ids[0]) :],
        # skip_special_tokens=True,
    )
    print(generated_text)


@torch.inference_mode
def predict_stream(gpu_prop, thinking):
    # Load model
    MODEL_PATH = os.path.join(HF_MODEL_DIR, os.getenv("LLM_MODEL"))
    processor = AutoProcessor.from_pretrained(MODEL_PATH, use_fast=True, trust_remote_code=True)
    model = AutoModel.from_pretrained(
        MODEL_PATH,
        torch_dtype=torch.bfloat16,
        _attn_implementation="flash_attention_2" if gpu_prop.major >= 8 else None,
        trust_remote_code=True,
    ).to("cuda")

    model = model.eval()
    # print(f"{model.config=}")
    # print(f"{processor=}")
    # print(f"{model=}")

    # Construct prompt
    # text = "请用中文描述图片内容"
    text = f"请用中文描述图片内容，不要使用特殊字符回复。"
    # text = "Please reply to my message in Chinese simplified(简体中文), don't use Markdown format. 描述下图片内容"
    if thinking is True:
        text += "/think"
    if thinking is False:
        text += "/no_think"

    messages = [
        {
            "role": "user",
            "content": [
                {
                    "type": "image",
                    "url": "https://huggingface.co/datasets/huggingface/documentation-images/resolve/main/bee.jpg",
                },
                {"type": "text", "text": text},
            ],
        },
    ]

    inputs = processor.apply_chat_template(
        messages,
        add_generation_prompt=True,
        tokenize=True,
        return_dict=True,
        return_tensors="pt",
    ).to(model.device, dtype=torch.bfloat16)
    for key, value in inputs.items():
        print(f"{key}: {value.shape=}")
    input_ids = inputs["input_ids"]
    prompt = processor.decode(input_ids[0])
    print(f"{prompt=}")

    for i in range(3):
        print("---" * 20)
        streamer = TextIteratorStreamer(
            tokenizer=processor, skip_prompt=True, skip_special_tokens=True
        )
        # don't to do sampling
        generation_kwargs = dict(
            **inputs,
            do_sample=False,
            # do_sample=True,
            # temperature=0.2,
            # top_p=None,
            # num_beams=1,
            # repetition_penalty=1.5,
            max_new_tokens=1024,
            use_cache=True,
            streamer=streamer,
        )
        thread = Thread(target=model.generate, kwargs=generation_kwargs)
        thread.start()
        generated_text = ""
        start = perf_counter()
        times = []
        is_output_think = os.getenv("IS_OUTPUT_THINK", "1") == "1"
        for new_text in streamer:
            times.append(perf_counter() - start)
            print(new_text, end="", flush=True)
            if is_output_think is False:
                if "</think>" in new_text:
                    new_text = new_text.replace("</think>", "").strip("\n")
                    is_output_think = True
                else:
                    continue
            generated_text += new_text
            start = perf_counter()
        print(f"\n{i}. {generated_text=} TTFT: {times[0]:.2f}s total time: {sum(times):.2f}s")


@torch.inference_mode()
def chat(gpu_prop, thinking):
    # Load model
    model_name = os.getenv("LLM_MODEL")
    MODEL_PATH = os.path.join(HF_MODEL_DIR, model_name)
    processor = AutoProcessor.from_pretrained(MODEL_PATH, use_fast=True, trust_remote_code=True)
    model = AutoModel.from_pretrained(
        MODEL_PATH,
        torch_dtype=torch.bfloat16,
        _attn_implementation="flash_attention_2" if gpu_prop.major >= 8 else None,
        trust_remote_code=True,
    ).to("cuda")

    model = model.eval()
    # print(f"{model.config=}")
    # print(f"{processor=}")
    # print_model_params(model, f"{model_name}")

    # Construct history chat messages
    text = "讲一个故事"
    if thinking is True:
        text += "/think"
    if thinking is False:
        text += "/no_think"
    messages = [
        {
            "role": "system",
            "content": [
                {
                    "type": "text",
                    "text": "你是一个图片描述助手，你可以用中文描述图片内容，不要使用特殊字符回复。",
                }
            ],
        },
        {
            "role": "user",
            "content": [
                {
                    "type": "image",
                    "url": "https://huggingface.co/datasets/huggingface/documentation-images/resolve/main/bee.jpg",
                },
                {"type": "text", "text": "请描述图片内容"},
            ],
        },
        {
            "role": "assistant",
            "content": [
                {
                    "type": "text",
                    "text": "画面中，几朵粉色的花朵在绿意环绕的环境中绽放，其中一朵花上停着一只蜜蜂，蜜蜂正忙碌地在花蕊间采蜜。花朵有的盛开，花瓣舒展，花蕊明黄；有的已凋谢，花瓣枯萎卷曲。画面中还能看到红色的花朵点缀其间，背景是模糊的绿色植物与草地，整体呈现出自然花园里生机与凋零交织的景象，色彩柔和，充满生活气息。",  # remove think content
                }
            ],
        },
        {
            "role": "user",
            "content": [
                {
                    "type": "image",
                    "url": "https://huggingface.co/datasets/huggingface/documentation-images/resolve/main/bee.jpg",
                },
                {"type": "text", "text": "图片中有几只蜜蜂"},
            ],
        },
        {
            "role": "assistant",
            "content": [
                {
                    "type": "text",
                    "text": "图片中有一只蜜蜂。",  # remove think content
                }
            ],
        },
        {
            "role": "user",
            "content": [
                {
                    "type": "image",
                    "url": "https://huggingface.co/datasets/huggingface/documentation-images/resolve/main/bee.jpg",
                },
                {"type": "text", "text": text},
            ],
        },
    ]

    inputs = processor.apply_chat_template(
        messages,
        add_generation_prompt=True,
        tokenize=True,
        return_dict=True,
        return_tensors="pt",
    ).to(model.device, dtype=torch.bfloat16)
    for key, value in inputs.items():
        print(f"{key}: {value.shape=}")
    input_ids = inputs["input_ids"]
    prompt = processor.decode(input_ids[0])
    print(f"{prompt=}")

    generated_ids = model.generate(
        **inputs,
        do_sample=False,
        # repetition_penalty=1.5,
        max_new_tokens=2048,
    )
    print(f"{generated_ids.shape=}")
    generated_text = processor.decode(
        generated_ids[0][len(input_ids[0]) :],
        # skip_special_tokens=True,
    )
    print(generated_text)


@torch.inference_mode()
def text_vision_chat(gpu_prop, thinking):
    # Load model
    model_name = os.getenv("LLM_MODEL")
    MODEL_PATH = os.path.join(HF_MODEL_DIR, model_name)
    processor = AutoProcessor.from_pretrained(MODEL_PATH, use_fast=True, trust_remote_code=True)
    model = AutoModel.from_pretrained(
        MODEL_PATH,
        torch_dtype=torch.bfloat16,
        _attn_implementation="flash_attention_2" if gpu_prop.major >= 8 else None,
        trust_remote_code=True,
    ).to("cuda")

    model = model.eval()
    # print(f"{model.config=}")
    # print(f"{processor=}")
    # print_model_params(model, f"{model_name}")

    # Construct history chat messages
    text = "讲一个故事"
    if thinking is True:
        text += "/think"
    if thinking is False:
        text += "/no_think"
    messages = [
        {
            "role": "system",
            "content": [
                {
                    "type": "text",
                    "text": "你是一个图片描述助手，你可以用中文描述图片内容，不要使用特殊字符回复。",
                }
            ],
        },
        {
            "role": "user",
            "content": [
                {"type": "text", "text": "你叫什么名字"},
            ],
        },
        {
            "role": "assistant",
            "content": [
                {
                    "type": "text",
                    "text": "我叫Kwai Keye，是由快手基础大模型团队打造的多模态大模型。我基于海量数据和知识持续进化，拥有广泛的多领域知识和创作才能，具备出色的视觉感知理解、语言理解和生成能力，能够理解并高效执行各类任务。我擅长图像问答、视频问答、知识问答、文案创作、文字翻译、数学逻辑、代码理解和编写等任务，虽然我现在并不完美，偶尔也会出一些小差错，但我仍然在努力提升我的能力和准确度，期待能够为你提供更智能、轻快的互动体验。",  # remove analysis/think content
                }
            ],
        },
        {
            "role": "user",
            "content": [
                {
                    "type": "image",
                    "url": "https://huggingface.co/datasets/huggingface/documentation-images/resolve/main/bee.jpg",
                },
                {"type": "text", "text": "图片中有几只蜜蜂"},
            ],
        },
        {
            "role": "assistant",
            "content": [
                {
                    "type": "text",
                    "text": "图片中有一只蜜蜂。",  # remove analysis/think content
                }
            ],
        },
        {
            "role": "user",
            "content": [
                {"type": "text", "text": text},
            ],
        },
    ]

    round = int(os.getenv("ROUND", "1")) * 2 if int(os.getenv("ROUND", "1")) > 0 else 2
    inputs = processor.apply_chat_template(
        messages[:round],
        add_generation_prompt=True,
        tokenize=True,
        return_dict=True,
        return_tensors="pt",
    ).to(model.device, dtype=torch.bfloat16)
    for key, value in inputs.items():
        print(f"{key}: {value.shape=}")
    input_ids = inputs["input_ids"]
    prompt = processor.decode(input_ids[0])
    print(f"{prompt=}")

    generated_ids = model.generate(
        **inputs,
        do_sample=False,
        # repetition_penalty=1.5,
        max_new_tokens=2048,
    )
    print(f"{generated_ids.shape=}")
    generated_text = processor.decode(
        generated_ids[0][len(input_ids[0]) :],
        # skip_special_tokens=True,
    )
    print(generated_text)


@torch.inference_mode()
def text_video_chat(gpu_prop, thinking):
    # Load model
    model_name = os.getenv("LLM_MODEL")
    MODEL_PATH = os.path.join(HF_MODEL_DIR, model_name)
    processor = AutoProcessor.from_pretrained(MODEL_PATH, use_fast=True, trust_remote_code=True)
    model = AutoModel.from_pretrained(
        MODEL_PATH,
        torch_dtype=torch.bfloat16,
        _attn_implementation="flash_attention_2" if gpu_prop.major >= 8 else None,
        trust_remote_code=True,
    ).to("cuda")

    model = model.eval()
    # print(f"{model.config=}")
    # print(f"{processor=}")
    # print_model_params(model, f"{model_name}")

    # from meigen-multitalk generated video
    video_file = os.path.join(VIDEO_OUTPUT_DIR, "multi_long_exp.mp4")
    print(video_file)

    # Construct history chat messages
    text = "讲一个故事"
    if thinking is True:
        text += "/think"
    if thinking is False:
        text += "/no_think"
    messages = [
        {
            "role": "system",
            "content": [
                {
                    "type": "text",
                    "text": "你是一个视频描述助手，你可以用中文描述视频,图片内容，不要使用特殊字符回复。",
                }
            ],
        },
        {
            "role": "user",
            "content": [
                {"type": "text", "text": "你叫什么名字"},
            ],
        },
        {
            "role": "assistant",
            "content": [
                {
                    "type": "text",
                    "text": "我叫Kwai Keye，是由快手基础大模型团队打造的多模态大模型。我基于海量数据和知识持续进化，拥有广泛的多领域知识和创作才能，具备出色的视觉感知理解、语言理解和生成能力，能够理解并高效执行各类任务。我擅长图像问答、视频问答、知识问答、文案创作、文字翻译、数学逻辑、代码理解和编写等任务，虽然我现在并不完美，偶尔也会出一些小差错，但我仍然在努力提升我的能力和准确度，期待能够为你提供更智能、轻快的互动体验。",  # remove analysis/think content
                }
            ],
        },
        {
            "role": "user",
            "content": [
                {
                    "type": "image",
                    "image": "https://huggingface.co/datasets/huggingface/documentation-images/resolve/main/bee.jpg",
                },
                {"type": "text", "text": "图片中有几只蜜蜂"},
            ],
        },
        {
            "role": "assistant",
            "content": [
                {
                    "type": "text",
                    "text": "图片中有一只蜜蜂。",  # remove analysis/think content
                }
            ],
        },
        {
            "role": "user",
            "content": [
                {
                    "type": "video",
                    "video": video_file,
                    # "max_pixels": 480 * 480,
                    # "fps": 1.0,
                },
                {"type": "text", "text": "描述这个视频"},
            ],
        },
        {
            "role": "assistant",
            "content": [
                {
                    "type": "text",
                    "text": "视频中展示了一对男女在录音室内亲密互动的场景。男子穿着浅色衬衫，女子则身着带有花纹的无袖连衣裙，两人面对面站立，彼此靠近，似乎在深情对唱或交流。他们面前各有一个专业麦克风，背景是典型的录音室环境，有隔音板和声学处理设施。整个画面营造出一种温馨且专注的氛围，显示出两人之间深厚的情感联系。",  # remove analysis/think content
                }
            ],
        },
        {
            "role": "user",
            "content": [
                {"type": "text", "text": text},
            ],
        },
    ]

    round = int(os.getenv("ROUND", "1")) * 2 if int(os.getenv("ROUND", "1")) > 0 else 2
    # In Keye-VL, frame rate information is also input into the model to align with absolute time.
    text = processor.apply_chat_template(
        messages[:round],
        add_generation_prompt=True,
        tokenize=False,
        # return_dict=True,
        # return_tensors="pt",
    )
    print(text)

    image_inputs, video_inputs, video_kwargs = process_vision_info(
        messages[:round], return_video_kwargs=True
    )
    print(image_inputs, video_inputs, video_kwargs)
    inputs = processor(
        text=[text],
        images=image_inputs,
        videos=video_inputs,
        padding=True,
        return_tensors="pt",
        **video_kwargs,
    )
    inputs = inputs.to(model.device, dtype=torch.bfloat16)

    for key, value in inputs.items():
        print(f"{key}: {value.shape=}")
    input_ids = inputs["input_ids"]
    prompt = processor.decode(input_ids[0])
    print(f"{prompt=}")

    generated_ids = model.generate(
        **inputs,
        do_sample=False,
        # repetition_penalty=1.5,
        max_new_tokens=2048,
    )
    print(f"{generated_ids.shape=}")
    generated_text = processor.decode(
        generated_ids[0][len(input_ids[0]) :],
        # skip_special_tokens=True,
    )
    print(generated_text)


@torch.inference_mode()
def chat_tool(gpu_prop, thinking):
    # Load model
    model_name = os.getenv("LLM_MODEL")
    MODEL_PATH = os.path.join(HF_MODEL_DIR, model_name)
    processor = AutoProcessor.from_pretrained(MODEL_PATH, use_fast=True, trust_remote_code=True)
    model = AutoModel.from_pretrained(
        MODEL_PATH,
        torch_dtype=torch.bfloat16,
        _attn_implementation="flash_attention_2" if gpu_prop.major >= 8 else None,
        trust_remote_code=True,
    ).to("cuda")

    model = model.eval()

    tools = [
        {
            "type": "function",
            "function": {
                "name": "get_weather",
                "description": "Get weather of an location, the user shoud supply a location first",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "location": {
                            "type": "string",
                            "description": "The city and state, e.g. San Francisco, CA",
                        }
                    },
                    "required": ["location"],
                },
            },
        },
    ]

    text = "北京的天气"
    if thinking is None:
        text += "/agentic_think"
    if thinking is True:
        text += "/think"
    if thinking is False:
        text += "/no_think"
    messages = [
        {
            "role": "system",
            "content": [
                {
                    "type": "text",
                    "text": "你是一个智能助手，不要使用特殊字符回复。"
                    "\n 提供的工具如下：\n"
                    "\n Tool: get_weather \n"
                    "\n Description: Get weather of an location, the user shoud supply a location first \n"
                    "\n Arguments: location\n\n"
                    "\n 根据用户的问题选择合适的工具。如果不需要工具，直接回复。\n",
                }
            ],
        },
        {
            "role": "user",
            "content": [
                # {
                #    "type": "image",
                #    "url": "https://huggingface.co/datasets/huggingface/documentation-images/resolve/main/bee.jpg",
                # },
                {"type": "text", "text": text},
            ],
        },
    ]

    inputs = processor.apply_chat_template(
        messages,
        tools=tools,
        add_generation_prompt=True,
        tokenize=True,
        return_dict=True,
        return_tensors="pt",
    ).to(model.device, dtype=torch.bfloat16)
    for key, value in inputs.items():
        print(f"{key}: {value.shape=}")
    input_ids = inputs["input_ids"]
    prompt = processor.decode(input_ids[0])
    print(f"{prompt=}")

    temperature = 0.6
    generated_ids = model.generate(
        **inputs,
        do_sample=True if temperature > 0 else False,
        temperature=temperature,
        top_k=20,
        top_p=0.9,
        repetition_penalty=1.1,
        max_new_tokens=1024,
    )
    print(f"{generated_ids.shape=}")
    generated_text = processor.decode(
        generated_ids[0][len(input_ids[0]) :],
        # skip_special_tokens=True,
    )
    print(generated_text)


@torch.inference_mode()
def chat_json_mode(gpu_prop, thinking):
    # Load model
    model_name = os.getenv("LLM_MODEL")
    MODEL_PATH = os.path.join(HF_MODEL_DIR, model_name)
    processor = AutoProcessor.from_pretrained(MODEL_PATH, use_fast=True, trust_remote_code=True)
    model = AutoModel.from_pretrained(
        MODEL_PATH,
        torch_dtype=torch.bfloat16,
        _attn_implementation="flash_attention_2" if gpu_prop.major >= 8 else None,
        trust_remote_code=True,
    ).to("cuda")

    model = model.eval()

    system_prompt = """
    The user will provide some exam text. Please parse the "question" and "answer" and output them in JSON format. 
    
    EXAMPLE INPUT: 
    Which is the highest mountain in the world? Mount Everest.
    
    EXAMPLE JSON OUTPUT:
    {
        "question": "Which is the highest mountain in the world?",
        "answer": "Mount Everest"
    }
    """
    text = "Which is the longest river in the world? The Nile River."
    if thinking is True:
        text += "/think"
    if thinking is False:
        text += "/no_think"
    messages = [
        {
            "role": "system",
            "content": [
                {
                    "type": "text",
                    "text": system_prompt,
                }
            ],
        },
        {
            "role": "user",
            "content": [
                # {
                #    "type": "image",
                #    "url": "https://huggingface.co/datasets/huggingface/documentation-images/resolve/main/bee.jpg",
                # },
                {
                    "type": "text",
                    "text": text,
                },
            ],
        },
    ]

    inputs = processor.apply_chat_template(
        messages,
        add_generation_prompt=True,
        tokenize=True,
        return_dict=True,
        return_tensors="pt",
    ).to(model.device, dtype=torch.bfloat16)
    for key, value in inputs.items():
        print(f"{key}: {value.shape=}")
    input_ids = inputs["input_ids"]
    prompt = processor.decode(input_ids[0])
    print(f"{prompt=}")

    temperature = 0.6
    generated_ids = model.generate(
        **inputs,
        do_sample=True if temperature > 0 else False,
        temperature=temperature,
        top_k=20,
        top_p=0.9,
        repetition_penalty=1.1,
        max_new_tokens=1024,
    )
    print(f"{generated_ids.shape=}")
    generated_text = processor.decode(
        generated_ids[0][len(input_ids[0]) :],
        # skip_special_tokens=True,
    )
    print(generated_text)


"""
https://huggingface.co/Kwai-Keye/Keye-VL-8B-Preview
Following Qwen3, keye also offer a soft switch mechanism
- nothing(default): auto analysis model 
    - no-thinking: (<analysis>***</analysis>xxxxx)
    - thinking: (<analysis>***</analysis><think>***</think><answer>xxxxx</answer>)
- /think: thinking model (<think>***</think><answer>xxxxx</answer>)
- /no_think: no thinking model (xxxxx)

# download model
modal run src/download_models.py --repo-ids "Kwai-Keye/Keye-VL-8B-Preview"

# dump model
IMAGE_GPU=L4 modal run src/llm/transformers/vlm/keye.py --task dump_model

# text model
IMAGE_GPU=L4 modal run src/llm/transformers/vlm/keye.py --task predict_text
IMAGE_GPU=L4 modal run src/llm/transformers/vlm/keye.py --task predict_text --thinking
IMAGE_GPU=L4 modal run src/llm/transformers/vlm/keye.py --task predict_text --no-thinking
IMAGE_GPU=L4 modal run src/llm/transformers/vlm/keye.py --task chat_text
IMAGE_GPU=L4 modal run src/llm/transformers/vlm/keye.py --task chat_text --thinking
IMAGE_GPU=L4 modal run src/llm/transformers/vlm/keye.py --task chat_text --no-thinking

# vision text/image chat
IMAGE_GPU=L4 modal run src/llm/transformers/vlm/keye.py --task predict 
IMAGE_GPU=L4 modal run src/llm/transformers/vlm/keye.py --task predict --no-thinking
IMAGE_GPU=L4 modal run src/llm/transformers/vlm/keye.py --task predict --thinking
IMAGE_GPU=L4 modal run src/llm/transformers/vlm/keye.py --task predict_stream
IMAGE_GPU=L4 modal run src/llm/transformers/vlm/keye.py --task predict_stream --no-thinking
IMAGE_GPU=L4 modal run src/llm/transformers/vlm/keye.py --task predict_stream --thinking

IMAGE_GPU=L4 modal run src/llm/transformers/vlm/keye.py --task chat 
IMAGE_GPU=L4 modal run src/llm/transformers/vlm/keye.py --task chat --no-thinking
IMAGE_GPU=L4 modal run src/llm/transformers/vlm/keye.py --task chat --thinking

IMAGE_GPU=L4 modal run src/llm/transformers/vlm/keye.py --task text_vision_chat
ROUND=2 IMAGE_GPU=L4 modal run src/llm/transformers/vlm/keye.py --task text_vision_chat
ROUND=3 IMAGE_GPU=L4 modal run src/llm/transformers/vlm/keye.py --task text_vision_chat

# vision text/image/video chat
ROUND=1 IMAGE_GPU=L4 modal run src/llm/transformers/vlm/keye.py --task text_video_chat
ROUND=2 IMAGE_GPU=L4 modal run src/llm/transformers/vlm/keye.py --task text_video_chat
ROUND=3 IMAGE_GPU=L4 modal run src/llm/transformers/vlm/keye.py --task text_video_chat
ROUND=4 IMAGE_GPU=L4 modal run src/llm/transformers/vlm/keye.py --task text_video_chat

# 不支持funciton_calling 需要微调
IMAGE_GPU=L4 modal run src/llm/transformers/vlm/keye.py --task chat_tool
IMAGE_GPU=L4 modal run src/llm/transformers/vlm/keye.py --task chat_tool --thinking
IMAGE_GPU=L4 modal run src/llm/transformers/vlm/keye.py --task chat_tool --no-thinking

# json_mode支持, 但是输出markdown格式 ```json\n{}\n```
IMAGE_GPU=L4 modal run src/llm/transformers/vlm/keye.py --task chat_json_mode
"""


@app.local_entrypoint()
def main(task: str = "dump_model", thinking: Optional[bool] = None):
    print(task, thinking)
    tasks = {
        "dump_model": dump_model,
        "predict_text": predict_text,
        "chat_text": chat_text,
        "predict": predict,
        "predict_stream": predict_stream,
        "chat": chat,
        "text_vision_chat": text_vision_chat,
        "text_video_chat": text_video_chat,
        "chat_tool": chat_tool,
        "chat_json_mode": chat_json_mode,
    }
    if task not in tasks:
        raise ValueError(f"task {task} not found")
    print(f"running task {task}")
    run.remote(tasks[task], thinking)



import logging
from typing import cast

from fastapi import WebSocket
from apipeline.pipeline.pipeline import Pipeline
from apipeline.pipeline.task import PipelineParams, PipelineTask
from apipeline.pipeline.runner import PipelineRunner
from apipeline.frames import TextFrame, InterruptionTaskFrame, InterruptionFrame
from apipeline.processors.logger import FrameLogger

from src.cmd.bots.base_fastapi_websocket_server import AIFastapiWebsocketBot
from src.types.network.fastapi_websocket import FastapiWebsocketServerParams
from src.transports.fastapi_websocket_server import FastapiWebsocketTransport
from src.serializers.transcription_protobuf import TranscriptionFrameSerializer
from src.cmd.bots import register_ai_fastapi_ws_bots
from src.processors.a2a.a2a_live_processor import A2ALiveProcessor

from dotenv import load_dotenv

load_dotenv(override=True)


@register_ai_fastapi_ws_bots.register
class FastapiWebsocketA2ALiveBot(AIFastapiWebsocketBot):
    """
    use a2a live bot
    - no VAD
    - text/images/audio -> gemini live (BIDI) -> text/audio
    """

    def __init__(self, **args) -> None:
        super().__init__(**args)
        self.init_bot_config()

    async def arun(self):
        if self._websocket is None:
            return

        serializer = TranscriptionFrameSerializer()
        self.params = FastapiWebsocketServerParams(
            audio_in_enabled=True,
            audio_out_enabled=True,
            add_wav_header=True,
            vad_enabled=False,
            vad_audio_passthrough=True,
            serializer=serializer,
            audio_out_sample_rate=24000,
        )
        transport = FastapiWebsocketTransport(
            websocket=self._websocket,
            params=self.params,
        )
        self.a2a_processor: A2ALiveProcessor = cast(
            A2ALiveProcessor,
            self.get_a2a_processor(tag="a2a_live_processor"),
        )

        self.task = PipelineTask(
            Pipeline(
                [
                    transport.input_processor(),
                    FrameLogger(include_frame_types=[InterruptionTaskFrame, InterruptionFrame]),
                    self.a2a_processor,
                    transport.output_processor(),
                ]
            ),
            params=self._pipeline_params,
        )

        transport.add_event_handler("on_client_connected", self.on_client_connected)
        transport.add_event_handler("on_client_disconnected", self.on_client_disconnected)

        await PipelineRunner(handle_sigint=self._handle_sigint).run(self.task)

    async def on_client_connected(
        self,
        transport: FastapiWebsocketTransport,
        websocket: WebSocket,
    ):
        logging.info(f"on_client_connected client:{websocket.client}")
        client_id = f"{websocket.client.host}:{websocket.client.port}"
        self.session.set_client_id(client_id=client_id)
        self.a2a_processor.set_user_id(client_id)
        await self.a2a_processor.create_conversation()
        await self.a2a_processor.create_push_task()

        is_cn = self._bot_config.a2a and self._bot_config.a2a.language == "zh"
        user_hi_text = "请用中文介绍下自己。" if is_cn else "Please introduce yourself first."
        await self.task.queue_frame(TextFrame(text=user_hi_text))



import asyncio
import logging
import os
from typing import Any, Awaitable, Callable

from PIL import Image

from apipeline.frames.data_frames import TextFrame
from apipeline.frames.control_frames import EndFrame
from apipeline.pipeline.pipeline import Pipeline
from apipeline.pipeline.runner import PipelineRunner
from apipeline.pipeline.task import PipelineTask, PipelineParams
from apipeline.processors.aggregators.sentence import SentenceAggregator
from apipeline.processors.output_processor import OutputFrameProcessor
from apipeline.processors.logger import FrameLogger

from src.processors.image_capture_processor import ImageCaptureProcessor
from src.processors.aggregators.vision_image_frame import VisionImageFrameAggregator
from src.processors.llm.base import LLMProcessor
from src.processors.aggregators.openai_llm_context import (
    OpenAILLMContext,
    OpenAIAssistantContextAggregator,
    OpenAIUserContextAggregator,
)
from src.processors.speech.tts.tts_processor import TTSProcessor
from src.common.types import DailyParams
from src.cmd.bots.base_daily import DailyRoomBot
from src.transports.daily import DailyTransport
from src.types.frames.data_frames import (
    FunctionCallResultFrame,
    LLMMessagesFrame,
    UserImageRawFrame,
    VisionImageRawFrame,
)
from src.common.register import Register
from .. import register_ai_room_bots

register_tool_funtions = Register("daily-chat-vision-tool-functions")


@register_ai_room_bots.register
class DailyChatToolsVisionBot(DailyRoomBot):
    r"""
    use function tools llm model to chat (for text LLM)
    - when tool is describe image, use describe_image function which use vision model to describe the image with describe text (vision LLM)
        - just use vision lm(Vertical Domain(OCR, etc..) or General(VIT)) to detecte objects and describe the image/video with text
    - when tool is get weather, use get_weather function to get weahter info. this function just a demo

    - TODO: do more tool functions, e.g.: music gen/search, google search etc... become a agent
    !NOTE: need write system prompt to guide LLM to answer

    !THINKING: @weedge (gedigedix3) personal AI(agent)
    - if want a system engine become more easy, need train a MLLM(which can call tools with post training) to supporte e2e text,audio,vision (a big bang!);
    - if not, need base text LLM(which can call tools with post training) and more engine pipeline with tools(agents) to support ~!
    - if just develop app with using lm; prompt is all your need (zero or few shots), need know MLLM/LLM can do what, follow lm capability.
    - all tool function need to async run and set task timeout with using LLM's Parallel Tool
    - multi async agent sync msg with connector(local pipe; async queue or sync rpc)
    — Think about logically rigorous scenarios:
        - like o1 (need CoT prompting with llm, or use openai o1.)
    """

    def __init__(self, **args) -> None:
        super().__init__(**args)
        self.participant_uid = None  # support one user which chats with many bots
        self.max_function_call_cn = int(os.environ.get("MAX_FUNCTION_CALL_CN", "3"))
        self.vision_result = ""
        self.llm_context = OpenAILLMContext()
        self.init_bot_config()

    async def sink_out_cb(self, frame):
        """
        get vision processor result
        """
        if isinstance(frame, TextFrame):
            self.vision_result += frame.text

    @register_tool_funtions.register
    async def get_weather(
        self,
        function_name: str,
        tool_call_id: str,
        arguments: Any,
        llm: LLMProcessor,
        context: OpenAILLMContext,
        result_callback: Callable[[Any], Awaitable[None]],
    ):
        location = arguments["location"]
        logging.info(
            f"function_name:{function_name}, tool_call_id:{tool_call_id},"
            f"arguments:{arguments}, llm:{llm}, context:{context}"
        )
        # just a mock response
        # add result to assistant context
        self.get_weather_call_cn += 1
        if self.max_function_call_cn > self.get_weather_call_cn:
            await result_callback(f"The weather in {location} is currently 32 degrees and sunny.")
        else:
            self.get_weather_call_cn = 0

    @register_tool_funtions.register
    async def describe_image(
        self,
        function_name: str,
        tool_call_id: str,
        arguments: Any,
        llm: LLMProcessor,
        context: OpenAILLMContext,
        result_callback: Callable[[Any], Awaitable[None]],
    ):
        logging.info(
            f"function_name:{function_name}, tool_call_id:{tool_call_id},"
            f"arguments:{arguments}, llm:{llm}, context:{context}"
        )
        if "question" not in arguments:
            arguments["question"] = "describe image."
        images = self.image_capture_processor.capture_imgs.get(cls=list)
        if len(images) == 0:
            # no described image, so return tips
            await result_callback("no described image, please try again.")
            return

        image: UserImageRawFrame = images[0]
        frame = VisionImageRawFrame(
            text=arguments["question"],
            image=image.image,
            size=image.size,
            format=image.format,
            mode=image.mode,
        )
        vision_task = PipelineTask(
            Pipeline(
                [
                    # SentenceAggregator(),
                    # UserImageRequestProcessor(),
                    # VisionImageFrameAggregator(),
                    self.vision_llm_processor,
                    OutputFrameProcessor(cb=self.sink_out_cb),
                ]
            )
        )
        await vision_task.queue_frames(
            [
                frame,
                EndFrame(),
            ]
        )
        await PipelineRunner().run(vision_task)
        # return describe image tool result
        await result_callback(self.vision_result)

    async def arun(self):
        self.image_capture_processor = ImageCaptureProcessor()
        self.vision_llm_processor = self.get_vision_llm_processor()

        vad_analyzer = self.get_vad_analyzer()
        self.daily_params = DailyParams(
            audio_in_enabled=True,
            audio_out_enabled=True,
            vad_enabled=True,
            vad_analyzer=vad_analyzer,
            vad_audio_passthrough=True,
            transcription_enabled=False,
        )

        asr_processor = self.get_asr_processor()

        tts_processor: TTSProcessor = self.get_tts_processor()
        stream_info = tts_processor.get_stream_info()

        self.daily_params.audio_out_sample_rate = stream_info["sample_rate"]
        self.daily_params.audio_out_channels = stream_info["channels"]
        transport = DailyTransport(
            self.args.room_url,
            self.args.token,
            self.args.bot_name,
            self.daily_params,
        )
        transport.add_event_handler("on_first_participant_joined", self.on_first_participant_joined)
        transport.add_event_handler("on_participant_left", self.on_participant_left)
        transport.add_event_handler("on_call_state_updated", self.on_call_state_updated)

        if self._bot_config.llm and self._bot_config.llm.messages:
            self.llm_context.set_messages(self._bot_config.llm.messages)
        if self._bot_config.llm and self._bot_config.llm.tools:
            self.llm_context.set_tools(self._bot_config.llm.tools)
        llm_user_ctx_aggr = OpenAIUserContextAggregator(self.llm_context)
        llm_assistant_ctx_aggr = OpenAIAssistantContextAggregator(llm_user_ctx_aggr)
        llm_processor: LLMProcessor = self.get_llm_processor()

        # register function
        logging.info(f"register tool functions: {register_tool_funtions.items()}")
        llm_processor.register_function("get_weather", self.get_weather)
        self.get_weather_call_cn = 0
        llm_processor.register_function("describe_image", self.describe_image)
        self.describe_image_call_cn = 0

        pipeline = Pipeline(
            [
                transport.input_processor(),
                # FrameLogger(include_frame_types=[UserImageRawFrame]),
                self.image_capture_processor,
                asr_processor,
                llm_user_ctx_aggr,
                llm_processor,
                tts_processor,
                transport.output_processor(),
                # FrameLogger(include_frame_types=[FunctionCallResultFrame]),
                llm_assistant_ctx_aggr,
            ]
        )
        self.task = PipelineTask(pipeline, params=PipelineParams(enable_metrics=False))
        await PipelineRunner().run(self.task)

    async def on_first_participant_joined(self, transport: DailyTransport, participant):
        self.participant_uid = participant["id"]
        transport.capture_participant_video(participant["id"])
        if self.daily_params.transcription_enabled:
            transport.capture_participant_transcription(participant["id"])

        # joined use tts say "hello" to introduce with llm generate
        if (
            self._bot_config.tts
            and self._bot_config.llm
            and self._bot_config.llm.messages
            and len(self._bot_config.llm.messages) == 1
        ):
            hi_text = "Please introduce yourself first."
            if self._bot_config.llm.language and self._bot_config.llm.language == "zh":
                hi_text = "请用中文介绍下自己。"
            self._bot_config.llm.messages.append(
                {
                    "role": "user",
                    "content": hi_text,
                }
            )
            await self.task.queue_frames([LLMMessagesFrame(self._bot_config.llm.messages)])



#
# For licensing see accompanying LICENSE file.
# Copyright (C) 2025 Apple Inc. All Rights Reserved.
#
import torch
import torch.nn as nn
import torch.nn.functional as F

from transformers import CLIPImageProcessor
import llava.model.multimodal_encoder.mobileclip as mobileclip


class MobileCLIPVisionTower(nn.Module):
    def __init__(self, vision_tower, args, delay_load=False):
        super().__init__()

        self.is_loaded = False
        self.vision_tower_name = vision_tower
        self.tune_vision_tower = getattr(args, 'unfreeze_mm_vision_tower', False)
        self.input_image_size = int(vision_tower.split("_")[-1])

        # Delay load is disabled for now
        if not delay_load:
            self.load_model()
        elif getattr(args, 'unfreeze_mm_vision_tower', False):
            self.load_model()
        else:
            model_cfg = mobileclip.load_model_config(self.vision_tower_name)
            self.cfg_only = model_cfg

    def load_model(self, device_map=None):
        if self.is_loaded:
            print('{} is already loaded, `load_model` called again, skipping.'.format(self.vision_tower_name))
            return

        # Load model config
        model_cfg = mobileclip.load_model_config(self.vision_tower_name)

        # Override default image resolution
        model_cfg["image_cfg"]["image_size"] = self.input_image_size

        self.cfg_only = model_cfg

        # Build HF CLIPImageProcessor with MobileCLIP parameters
        self.image_processor = CLIPImageProcessor(crop_size={"height": model_cfg["image_cfg"]["image_size"],
                                                             "width": model_cfg["image_cfg"]["image_size"]},
                                                  image_mean=[0.0, 0.0, 0.0],
                                                  image_std=[1.0, 1.0, 1.0],
                                                  size={"shortest_edge": model_cfg["image_cfg"]["image_size"]})

        # Instantiate the image encoder
        self.vision_tower = mobileclip.MCi(model_name=model_cfg["image_cfg"]["model_name"],
                                           projection_dim=model_cfg["embed_dim"])

        if not self.tune_vision_tower:
            self.vision_tower.requires_grad_(False)

        self.is_loaded = True

    def feature_select(self, image_forward_outs):
        # Features from penultimate layer
        image_features = image_forward_outs["image_embeddings"]

        # Reshape 4D tensor to 3D
        B, C, H, W = image_features.shape
        image_features = image_features.reshape(B, C, H*W)
        image_features = image_features.transpose(1, 2)
        return image_features

    def forward(self, images):
        if self.tune_vision_tower:
            return self.forward_images(images)
        else:
            with torch.no_grad():
                return self.forward_images(images)

    def forward_images(self, images):
        if type(images) is list:
            image_features = []
            for image in images:
                image_forward_out = self.vision_tower(image.to(device=self.device, dtype=self.dtype).unsqueeze(0), return_image_embeddings=True)
                image_feature = self.feature_select(image_forward_out).to(image.dtype)
                image_features.append(image_feature)
        else:
            image_forward_outs = self.vision_tower(images.to(device=self.device, dtype=self.dtype), return_image_embeddings=True)
            image_features = self.feature_select(image_forward_outs).to(images.dtype)

        return image_features

    @property
    def dummy_feature(self):
        return torch.zeros(1, self.hidden_size, device=self.device, dtype=self.dtype)

    @property
    def dtype(self):
        return next(self.vision_tower.parameters()).dtype

    @property
    def device(self):
        return next(self.vision_tower.parameters()).device

    @property
    def config(self):
        return self.cfg_only

    @property
    def hidden_size(self):
        return self.config["image_cfg"]["embed_dim"]

    @property
    def num_patches_per_side(self):
        return self.config["image_cfg"]["image_size"] // self.config["image_cfg"]["patch_size"]

    @property
    def num_patches(self):
        return (self.config["image_cfg"]["image_size"] // self.config["image_cfg"]["patch_size"]) ** 2



import os
import asyncio
import logging

import unittest

from src.cmd.init import Env
from src.common.logger import Logger

r"""
python -m unittest test.cmd.test_init.TestEnv.test_save_to_yamls
"""


class TestEnv(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        Logger.init(os.getenv("LOG_LEVEL", "debug").upper(), is_file=False)
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_save_to_yamls(self):
        os.environ["RECORDER_TAG"] = "wakeword_rms_recorder"
        os.environ["CONF_ENV"] = "local"
        res = asyncio.run(Env.save_to_yamls())
        self.assertIsInstance(res, list)
        for file_path in res:
            print(file_path)
            self.assertIsNotNone(file_path)



import datetime
import logging
import logging.handlers
import os
import sys

import requests

from llava.constants import LOGDIR

server_error_msg = "**NETWORK ERROR DUE TO HIGH TRAFFIC. PLEASE REGENERATE OR REFRESH THIS PAGE.**"
moderation_msg = "YOUR INPUT VIOLATES OUR CONTENT MODERATION GUIDELINES. PLEASE TRY AGAIN."

handler = None


def build_logger(logger_name, logger_filename):
    global handler

    formatter = logging.Formatter(
        fmt="%(asctime)s | %(levelname)s | %(name)s | %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    # Set the format of root handlers
    if not logging.getLogger().handlers:
        logging.basicConfig(level=logging.INFO)
    logging.getLogger().handlers[0].setFormatter(formatter)

    # Redirect stdout and stderr to loggers
    stdout_logger = logging.getLogger("stdout")
    stdout_logger.setLevel(logging.INFO)
    sl = StreamToLogger(stdout_logger, logging.INFO)
    sys.stdout = sl

    stderr_logger = logging.getLogger("stderr")
    stderr_logger.setLevel(logging.ERROR)
    sl = StreamToLogger(stderr_logger, logging.ERROR)
    sys.stderr = sl

    # Get logger
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.INFO)

    # Add a file handler for all loggers
    if handler is None:
        os.makedirs(LOGDIR, exist_ok=True)
        filename = os.path.join(LOGDIR, logger_filename)
        handler = logging.handlers.TimedRotatingFileHandler(
            filename, when='D', utc=True, encoding='UTF-8')
        handler.setFormatter(formatter)

        for name, item in logging.root.manager.loggerDict.items():
            if isinstance(item, logging.Logger):
                item.addHandler(handler)

    return logger


class StreamToLogger(object):
    """
    Fake file-like stream object that redirects writes to a logger instance.
    """

    def __init__(self, logger, log_level=logging.INFO):
        self.terminal = sys.stdout
        self.logger = logger
        self.log_level = log_level
        self.linebuf = ''

    def __getattr__(self, attr):
        return getattr(self.terminal, attr)

    def write(self, buf):
        temp_linebuf = self.linebuf + buf
        self.linebuf = ''
        for line in temp_linebuf.splitlines(True):
            # From the io.TextIOWrapper docs:
            #   On output, if newline is None, any '\n' characters written
            #   are translated to the system default line separator.
            # By default sys.stdout.write() expects '\n' newlines and then
            # translates them so this is still cross platform.
            if line[-1] == '\n':
                self.logger.log(self.log_level, line.rstrip())
            else:
                self.linebuf += line

    def flush(self):
        if self.linebuf != '':
            self.logger.log(self.log_level, self.linebuf.rstrip())
        self.linebuf = ''


def disable_torch_init():
    """
    Disable the redundant torch default initialization to accelerate model creation.
    """
    import torch
    setattr(torch.nn.Linear, "reset_parameters", lambda self: None)
    setattr(torch.nn.LayerNorm, "reset_parameters", lambda self: None)


def violates_moderation(text):
    """
    Check whether the text violates OpenAI moderation API.
    """
    url = "https://api.openai.com/v1/moderations"
    headers = {"Content-Type": "application/json",
               "Authorization": "Bearer " + os.environ["OPENAI_API_KEY"]}
    text = text.replace("\n", "")
    data = "{" + '"input": ' + f'"{text}"' + "}"
    data = data.encode("utf-8")
    try:
        ret = requests.post(url, headers=headers, data=data, timeout=5)
        flagged = ret.json()["results"][0]["flagged"]
    except requests.exceptions.RequestException as e:
        flagged = False
    except KeyError as e:
        flagged = False

    return flagged


def pretty_print_semaphore(semaphore):
    if semaphore is None:
        return "None"
    return f"Semaphore(value={semaphore._value}, locked={semaphore.locked()})"



import os
import re
import modal

app = modal.App("download_datasets")

# We also define the dependencies for our Function by specifying an
# [Image](https://modal.com/docs/guide/images).

download_image = (
    modal.Image.debian_slim(python_version="3.11").apt_install("curl").pip_install("datasets")
)

HF_DATASET_DIR = "/datasets"
hf_dataset_vol = modal.Volume.from_name("datasets", create_if_missing=True)


@app.function(
    # gpu="T4",
    retries=0,
    cpu=8.0,
    image=download_image,
    # secrets=[modal.Secret.from_name("achatbot")],
    volumes={HF_DATASET_DIR: hf_dataset_vol},
    timeout=1200,
)
def process():
    import datasets

    data_source = "openai/gsm8k"
    data_source_path = os.path.join(HF_DATASET_DIR, data_source)

    dataset = datasets.load_dataset(data_source_path, "main")

    train_dataset = dataset["train"]
    print("raw train dataset", train_dataset[0])
    test_dataset = dataset["test"]
    print("raw test dataset", test_dataset[0])

    instruction_following = 'Let\'s think step by step and output the final answer after "####".'

    # add a row to each data item that represents a unique id
    def make_map_fn(split):
        def process_fn(example, idx):
            question_raw = example.pop("question")

            question = question_raw + " " + instruction_following

            answer_raw = example.pop("answer")
            solution = extract_solution(answer_raw)
            data = {
                "data_source": data_source,
                "prompt": [
                    {
                        "role": "user",
                        "content": question,
                    }
                ],
                "ability": "math",
                "reward_model": {"style": "rule", "ground_truth": solution},
                "extra_info": {
                    "split": split,
                    "index": idx,
                    "answer": answer_raw,
                    "question": question_raw,
                },
            }
            return data

        return process_fn

    train_dataset = train_dataset.map(function=make_map_fn("train"), with_indices=True)
    print(train_dataset[0],flush=True)
    test_dataset = test_dataset.map(function=make_map_fn("test"), with_indices=True)
    print(test_dataset[0],flush=True)

    local_dir = data_source_path

    train_dataset.to_parquet(os.path.join(local_dir, "train.parquet"))
    test_dataset.to_parquet(os.path.join(local_dir, "test.parquet"))


def extract_solution(solution_str):
    solution = re.search("#### (\\-?[0-9\\.\\,]+)", solution_str)
    assert solution is not None
    final_solution = solution.group(0)
    final_solution = final_solution.split("#### ")[1].replace(",", "")
    # print(final_solution)
    return final_solution


"""
modal run src/train/verl/examples/data_preprocess/gsm8k.py
"""


@app.local_entrypoint()
def main():
    process.remote()



import click
import uvicorn

from a2a.server.agent_execution import AgentExecutor
from a2a.server.apps import A2AStarletteApplication
from a2a.server.request_handlers.default_request_handler import (
    DefaultRequestHandler,
)
from a2a.server.tasks.inmemory_task_store import InMemoryTaskStore
from a2a.types import (
    AgentCapabilities,
    AgentCard,
    AgentSkill,
    GetTaskRequest,
    GetTaskResponse,
    SendMessageRequest,
    SendMessageResponse,
)

from .agent_executor import GitMcpAgentExecutor


class A2ARequestHandler(DefaultRequestHandler):
    """A2A Request Handler for the achatbot Repo Agent."""

    def __init__(self, agent_executor: AgentExecutor, task_store: InMemoryTaskStore):
        super().__init__(agent_executor, task_store)

    async def on_get_task(self, request: GetTaskRequest) -> GetTaskResponse:
        return await super().on_get_task(request)

    async def on_message_send(self, request: SendMessageRequest) -> SendMessageResponse:
        return await super().on_message_send(request)


@click.command()
@click.option("--host", "host", default="localhost")
@click.option("--port", "port", default=6666)
def main(host: str, port: int):
    """Start the achatbot Repo Agent server.

    This function initializes the achatbot Repo Agent server with the specified host and port.
    It creates an agent card with the agent's name, description, version, and capabilities.

    Args:
        host (str): The host address to run the server on.
        port (int): The port number to run the server on.
    """
    skill = AgentSkill(
        id="answer_detail_about_achatbot_repo",
        name="Answer any information about a chatbot repo",
        description="The agent will look up the information about a chatbot repo and answer the question.",
        tags=["achatbot repo"],
        examples=["What is a chatbot repo?"],
    )

    agent_card = AgentCard(
        name="achatbot Agent",
        description="a chatbot knowledge agent who has information about a chatbot project and can answer questions about it",
        url=f"http://{host}:{port}/",
        version="1.0.0",
        default_input_modes=["text"],
        default_output_modes=["text"],
        capabilities=AgentCapabilities(
            input_modes=["text"],
            output_modes=["text"],
            streaming=True,
        ),
        skills=[skill],
        # authentication=AgentAuthentication(schemes=['public']),
        examples=["What is achatbot?"],
    )

    task_store = InMemoryTaskStore()
    request_handler = A2ARequestHandler(
        agent_executor=GitMcpAgentExecutor(),
        task_store=task_store,
    )

    server = A2AStarletteApplication(agent_card=agent_card, http_handler=request_handler)
    uvicorn.run(server.build(), host=host, port=port)


"""
python -m a2a_server.cmd.mcp_github
"""
if __name__ == "__main__":
    main()
